import unittest
import os

from delta.core import DeltaCore
from delta.vcs.handler import DeltaGitHandler


class FakeRun:
    def configure(**kwargs):
        pass

    def run(**kwargs):
        pass

    def monitor(**kwargs):
        pass

    def stop(**kwargs):
        pass

    def resume(**kwargs):
        pass


class TestCoreRun(unittest.TestCase):
    def test_run_start_initialized(self):
        with DeltaCore() as core:
            try:
                core._run = FakeRun()
                kwargs = {

                }
                core.run_start(**kwargs)
            except Exception:
                pass

    def test_run_stop_initialized(self):
        with DeltaCore() as core:
            try:
                core._run = FakeRun()
                kwargs = {

                }
                core.run_stop(**kwargs)
            except Exception:
                pass

    def test_run_resume_initialized(self):
        with DeltaCore() as core:
            try:
                core._run = FakeRun()
                kwargs = {

                }
                core.run_resume(**kwargs)
            except Exception:
                pass

    def test_run_monitor_initialized(self):
        with DeltaCore() as core:
            try:
                core._run = FakeRun()
                kwargs = {

                }
                core.run_monitor(**kwargs)
            except Exception:
                pass

    def test_run_configure_initialized(self):
        with DeltaCore() as core:
            core._run = FakeRun()
            try:
                kwargs = {

                }
                core.run_configure(**kwargs)
            except Exception:
                pass

    def test_run_start_not_initialized(self):
        with DeltaCore() as core:
            try:
                kwargs = {

                }
                core.run_start(**kwargs)
            except Exception:
                pass

    def test_run_stop_not_initialized(self):
        with DeltaCore() as core:
            try:
                kwargs = {

                }
                core.run_stop(**kwargs)
            except Exception:
                pass

    def test_run_resume_not_initialized(self):
        with DeltaCore() as core:
            try:
                kwargs = {

                }
                core.run_resume(**kwargs)
            except Exception:
                pass

    def test_run_monitor_not_initialized(self):
        with DeltaCore() as core:
            try:
                kwargs = {

                }
                core.run_monitor(**kwargs)
            except Exception:
                pass

    def test_run_configure_not_initialized(self):
        with DeltaCore() as core:
            try:
                kwargs = {

                }
                core.run_configure(**kwargs)
            except Exception:
                pass



"""Add parent_run column in run table

Revision ID: 24bd81ed7dad
Revises: 1ec1eabe45bf
Create Date: 2024-05-24 10:28:38.778180

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "24bd81ed7dad"
down_revision: Union[str, None] = "1ec1eabe45bf"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("runs", sa.Column("parent_run", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("runs", "parent_run")
    # ### end Alembic commands ###



from delta.manifest.parser import (
    Copyright, Input, License, Manifest, ManifestParser, Model, Output,
    Resource
)


class ManifestParserV1v0(ManifestParser):
    @classmethod
    def parse_license(cls, data: dict) -> License:
        return License(
            name=data["name"],
            url=data['url'],
            description=data.get("description"),
            copyrights=[Copyright(**e) for e in data["copyrights"]]
        )

    @classmethod
    def parse_resource(cls, data: dict) -> Resource:
        return Resource(
            name=data["name"],
            type="url",
            value=data["source_url"]
        )

    @classmethod
    def parse_input(cls, data: dict) -> Input:
        raise NotImplementedError

    @classmethod
    def parse_output(cls, data: dict) -> Output:
        raise NotImplementedError

    @classmethod
    def parse_model(cls, data: dict) -> Model:
        return Model(
            path=data["path"],
            type=data["type"],
            parameters=data["parameters"]
        )

    @classmethod
    def parse(cls, data: dict) -> Manifest:
        resources = {}
        if "resources" in data:
            for resource in data["resources"]["items"]:
                resources[resource['name']] = cls.parse_resource(resource)

        models = {}
        if "models" in data:
            models["single_model"] = cls.parse_model(data["models"])

        return Manifest(
            name=data["name"],
            description=data["description"],
            license=cls.parse_license(data["license"]),
            resources=resources,
            models=models
        )



"""add secret_value column to parameters

Revision ID: a2bfaf2e8406
Revises: 24bd81ed7dad
Create Date: 2024-07-25 11:43:52.547883

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'a2bfaf2e8406'
down_revision: Union[str, None] = '24bd81ed7dad'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('parameters', sa.Column('secret_value', sa.LargeBinary(), nullable=True))
    op.alter_column('runs', 'return_code',
               existing_type=sa.INTEGER(),
               nullable=False)
    op.alter_column('runs', 'message',
               existing_type=sa.VARCHAR(),
               nullable=False)
    op.alter_column('runs', 'parent_run',
               existing_type=sa.VARCHAR(),
               nullable=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('runs', 'parent_run',
               existing_type=sa.VARCHAR(),
               nullable=True)
    op.alter_column('runs', 'message',
               existing_type=sa.VARCHAR(),
               nullable=True)
    op.alter_column('runs', 'return_code',
               existing_type=sa.INTEGER(),
               nullable=True)
    op.drop_column('parameters', 'secret_value')
    # ### end Alembic commands ###



from typing import Any

from pydantic import BaseModel, Field


class Copyright(BaseModel):
    company: str = ""
    years: list[int] = [2024]


class License(BaseModel):
    name: str = "LGPLv3"
    url: str = "https://www.gnu.org/licenses/gpl-3.0.txt"
    description: str = ""
    copyrights: list[Copyright] = Field(default_factory=list)


class Parameter(BaseModel):
    name: str
    type: str
    description: str | None = None


class Resource(Parameter):
    value: Any


class Input(Parameter):
    value: Any | None = None


class InputModel(Input):
    prefix: str | None = None


class Output(Parameter):
    pass


class OutputModel(Output):
    glob: str


class Model(BaseModel):
    path: str
    type: str
    parameters: dict[str, Any] | None = None
    inputs: dict[str, InputModel] | None = None
    outputs: dict[str, OutputModel] | None = None


class Dependency(BaseModel):
    id: str
    version: str


class Manifest(BaseModel):
    name: str
    description: str = ""
    license: License = Field(default_factory=License)
    short_description: str | None = None
    owner: str
    resources: dict[str, Resource] | None = None
    inputs: dict[str, Input] | None = None
    outputs: dict[str, Output] | None = None
    models: dict[str, Model] | None = None
    dependencies: dict[str, Dependency] | None = None



import sys
import os

sys.path.append(os.path.join(os.path.dirname(__file__),
                             os.path.join('runners',
                                          os.path.join('resources',
                                                       'site-packages'))))



import os
import logging
from delta.store.interfaces import IData
from delta.store.interfaces import IStore
from delta.store.impl.drb_data_io_impl import DrbDataIOImpl
from delta.store.impl.fetcher_all_impl import FetcherAllImpl

from delta.manifest import manifest

logger = logging.getLogger('Delta')


class StoreImpl(IStore):

    def _write_on_manifest(self, name, path):
        """
        Write resource to manifest
        """
        if not os.path.exists('manifest.json'):
            manifest.write_manifest()
        _manifest = manifest.read_manifest()

        if "resources" not in _manifest:
            _manifest["resources"] = {
                "path": "resources",
                "items": []
            }

        items = _manifest["resources"]["items"]
        names = [item["name"] for item in items]

        if name not in names:
            json = {
                "name": name,
                "source_url": path
            }
            logger.info("Updating the manifest...")
            _manifest["resources"]["items"].append(json)
            manifest.write_manifest(content=_manifest)

    def add_resource(self, resources_path: str = './resources', **kwargs):
        """
        Add a resource to manifest and resources folder.
        So it will fetch data and write file to resources folder.

        :Keyword Arguments:
        * *name* (``str``) --
            Resource name.
        * *path* (``str``) --
            Data where we can find the resource file.
        * *download* (``boolean``) --
            A boolean that tell if the data will be downloaded
            in the resources folder.
        """
        data = IData()
        data_name = kwargs.get('name')
        data_path = kwargs.get('path')
        download = kwargs.get('download')
        data.add_info(name=data_name, path=data_path)

        if not os.path.exists(resources_path):
            try:
                os.makedirs(resources_path, exist_ok=True)
            except OSError:
                logger.warning(
                    f"WARNING: Failed to create {resources_path}...")

        if download:
            file_path = os.path.abspath(
                os.path.join(resources_path, data_name))
            with open(file_path, 'wb') as resource:
                drb_dataIO = DrbDataIOImpl(data, resource)
                fetcher = FetcherAllImpl(drb_dataIO)
                logger.info(f"Fetching {data_path}...")
                fetcher.fetch()
                logger.info(f"INFO: {data_path} has been fetched.")

        self._write_on_manifest(data_name, data_path)

    def remove_resource(self, **kwargs):
        """
        Remove a resource entry from the manifest.

        :Keyword Arguments:
        * *name* (``str``) --
            Resource name.
        """

        data_name = kwargs.get('name')

        if not os.path.exists('manifest.json'):
            logger.info("No manifest found")
            return

        _manifest = manifest.read_manifest()

        if "resources" not in _manifest:
            logger.info("INFO: No resource entry found")
            return

        for item in _manifest['resources']['items']:
            if item['name'] == data_name:
                _manifest['resources']['items'].remove(item)
                logger.info(
                    f"INFO: The resource {data_name} has been deleted.")

        manifest.write_manifest(content=_manifest)

    def get_resources(self):
        if not os.path.exists('manifest.json'):
            logger.info("No manifest found")
            return

        _manifest = manifest.read_manifest()

        if "resources" not in _manifest:
            logger.info("INFO: No resource entry found")
            return

        return _manifest['resources']['items']



import unittest
import os

from delta.vcs.handler.delta_git_handler import DeltaGitHandler, DeltaException


class TestCommands(unittest.TestCase):

    def test_commit_but_not_initialized(self):
        gh = DeltaGitHandler()
        with self.assertRaises(DeltaException):
            gh.commit('message')

    def test_add_but_not_initialized(self):
        gh = DeltaGitHandler()
        with self.assertRaises(DeltaException):
            gh.add(['file1', 'file2'])

    def test_pull_but_not_initialized(self):
        gh = DeltaGitHandler()
        with self.assertRaises(DeltaException):
            gh.pull('origin')

    def test_push_but_not_initialized(self):
        gh = DeltaGitHandler()
        with self.assertRaises(DeltaException):
            gh.push('origin')

    def test_fetch_but_not_initialized(self):
        gh = DeltaGitHandler()
        with self.assertRaises(DeltaException):
            gh.fetch('origin')

    def test_create_remote_but_not_initialized(self):
        gh = DeltaGitHandler()
        with self.assertRaises(DeltaException):
            gh.create_remote('origin', 'url')

    def test_status_but_not_initialized(self):
        gh = DeltaGitHandler()
        with self.assertRaises(DeltaException):
            gh.status()

    def test_checkout_but_not_initialized(self):
        gh = DeltaGitHandler()
        with self.assertRaises(DeltaException):
            gh.checkout('new_branch')

    def test_create_tag_but_not_initialized(self):
        gh = DeltaGitHandler()
        with self.assertRaises(DeltaException):
            gh.tag('v0.0.0', 'message')

    def test_push_tag_but_not_initialized(self):
        gh = DeltaGitHandler()
        with self.assertRaises(DeltaException):
            gh.push('origin', 'v0.0.0')

    def test_list_tag_but_not_initialized(self):
        gh = DeltaGitHandler()
        with self.assertRaises(DeltaException):
            gh.list_tag()

    def test_delete_tag_but_not_initialized(self):
        gh = DeltaGitHandler()
        with self.assertRaises(DeltaException):
            gh.remove_tag('v1.0.0')

    def test_commit_initialized(self):
        gh = DeltaGitHandler()
        self.assertIsNone(gh.create('test_commit_initialized'))
        first_head = gh._repo.head.commit.hexsha
        self.assertIsNone(gh.commit('Test'))
        second_head = gh._repo.head.commit.hexsha
        self.assertNotEqual(first_head, second_head)

    def test_add_initialized(self):
        gh = DeltaGitHandler()
        gh.create('test_add_initialized')
        open('test_add_initialized/test_add', 'w').close()
        open('test_add_initialized/resources/no_add', 'w').close()
        open('test_add_initialized/models/test_add2', 'w').close()
        res = gh.add(['test_add',
                      'resources/no_add',
                      'models/test_add2'])
        compare = [base_entry[3] for base_entry in res]
        self.assertEqual(compare, ['test_add', 'models/test_add2'])

    def test_fetch_initialized(self):
        gh = DeltaGitHandler()
        self.assertIsNone(gh.create('test_fetch_initialized'))
        with self.assertRaises(DeltaException):
            gh.fetch('origin')
        with self.assertRaises(DeltaException):
            gh.fetch('')

    def test_pull_initialized(self):
        gh = DeltaGitHandler()
        gh.create('test_pull_initialized')
        with self.assertRaises(DeltaException):
            gh.pull('origin')
        with self.assertRaises(DeltaException):
            gh.pull('')

    def test_push_initialized(self):
        gh = DeltaGitHandler()
        gh.create('test_push_initialized')
        with self.assertRaises(DeltaException):
            gh.push('origin')
        with self.assertRaises(DeltaException):
            gh.push('')

    def test_tag_initialized(self):
        gh = DeltaGitHandler()
        self.assertIsNone(gh.create('test_tag_initialized'))
        self.assertIsNone(gh.set_config('test', 'test@toto.org'))

        res = gh.tag('v1.0.0', 'First version')
        self.assertEqual('v1.0.0', res)

        with self.assertRaises(DeltaException):
            gh.tag('v1.0.0', 'First fake version')

        with self.assertRaises(DeltaException):
            gh.push('origin', 'v1.0.0')
        with self.assertRaises(DeltaException):
            gh.push('', 'v1.0.0')

        res = gh.tag('v1.0.1', 'New version')
        self.assertEqual(['v1.0.0', 'v1.0.1'], gh.list_tag())
        gh.remove_tag('v1.0.0')
        self.assertEqual(['v1.0.1'], gh.list_tag())

        with self.assertRaises(DeltaException):
            gh.remove_tag('v1.0.0')
        self.assertEqual(['v1.0.1'], gh.list_tag())
        gh.remove_tag('v1.0.1')
        self.assertEqual([], gh.list_tag())

    def test_checkout_initialized(self):
        gh = DeltaGitHandler()
        gh.create('test_checkout_initialized')
        gh.checkout('toto')
        self.assertEqual('toto', gh._repo.active_branch.name)
        gh.checkout('tata')
        self.assertEqual('tata', gh._repo.active_branch.name)
        gh.checkout('toto')
        self.assertEqual('toto', gh._repo.active_branch.name)

    def test_branch(self):
        gh = DeltaGitHandler()
        with self.assertRaises(DeltaException):
            gh.branch('toto')

        with self.assertRaises(DeltaException):
            gh.list_branches()

        gh.create('test_branch')
        branches = gh.list_branches()
        self.assertEqual(['master'], branches)

        with self.assertRaises(DeltaException):
            gh.remove_branch('toto')

        gh.branch('toto')
        self.assertNotEqual(gh._repo.active_branch.name, 'toto')
        branches = gh.list_branches()
        self.assertEqual(['master', 'toto'], branches)

        gh.remove_branch('toto')
        branches = gh.list_branches()
        self.assertEqual(['master'], branches)

        gh.checkout('abc')
        gh.checkout('toto')
        branches = gh.list_branches()
        self.assertEqual(['abc', 'master', 'toto'], branches)



import glob
import hashlib
import json
import logging
import os
import shutil
import urllib.request
import uuid
from abc import ABC, abstractmethod
from typing import IO, AnyStr, Dict, List

import boto3
import s3fs
import yaml
from aiohttp import ClientError

from delta.manifest.parser import Manifest, parse


class DeltaStorageManager(ABC):
    @abstractmethod
    def add_run_directory(self, run_id: uuid.UUID) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_full_path(self, path):
        raise NotImplementedError

    @abstractmethod
    def get_models_filepaths(self, name, version, model_path):
        raise NotImplementedError

    @abstractmethod
    def get_deltatwin_manifest(self, name: str, version: str) -> Manifest:
        raise NotImplementedError

    @abstractmethod
    def get_deltatwin_workflow(self, name: str, version: str) -> dict:
        raise NotImplementedError

    @abstractmethod
    def get_data(self, run_id: uuid.UUID, job_id: str, basename: str) -> IO:
        raise NotImplementedError

    @abstractmethod
    def set_data(
        self, run_id: uuid.UUID, job_id: str, basename: str, stream: IO
    ):
        raise NotImplementedError

    @abstractmethod
    def find_data(
        self, run_id: uuid.UUID, job_id: str, _glob: str
    ) -> List[AnyStr]:
        raise NotImplementedError

    def get_data_info(self, data_path) -> Dict:
        return NotImplementedError

    @abstractmethod
    def remove_run(self, run_id: uuid.UUID) -> None:
        raise NotImplementedError

    @abstractmethod
    def push_deltatwin(self, directory, name: str, version: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def remove_deltatwin(self, name: str, version: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def has_deltatwin(self, name: str, version: str) -> bool:
        raise NotImplementedError


class DeltaLocalStorageManager(DeltaStorageManager):
    DEFAULT_DIRECTORY = os.path.join(os.getenv("HOME"), ".delta")
    LOGGER = logging.getLogger("LocalStorageManager")

    def __init__(self, base_dir: str = None) -> None:
        self._base_dir = base_dir or self.DEFAULT_DIRECTORY

        if not os.path.isdir(self._base_dir):
            self.LOGGER.info("Creating base storage")
            os.makedirs(self._base_dir)

        self._deltatwin_dir = os.path.join(self._base_dir, "deltatwins")
        if not os.path.isdir(self._deltatwin_dir):
            self.LOGGER.info("Creating deltatwin storage")
            os.makedirs(self._deltatwin_dir)

        self._run_dir = os.path.join(self._base_dir, "runs")
        if not os.path.isdir(self._run_dir):
            self.LOGGER.info(f"Creating run storage")
            os.makedirs(self._run_dir)

    def get_full_path(self, path):
        return os.path.join(self._base_dir, path)

    def get_models_filepaths(self, name, version, model_path):
        return os.path.join(self._deltatwin_dir, name, version, model_path)

    def add_run_directory(self, run_id: uuid.UUID) -> None:
        path = os.path.join(self._run_dir, str(run_id))
        if not os.path.isdir(path):
            os.makedirs(path)

    def get_deltatwin_manifest(self, name: str, version: str) -> Manifest:
        path = os.path.join(
            self._deltatwin_dir, name, version, "manifest.json")
        if not os.path.isfile(path):
            raise FileNotFoundError(
                f"No manifest found for deltatwin: {name} version {version}")
        return parse(path)

    def get_deltatwin_workflow(self, name: str, version: str) -> dict:
        path = glob.glob(f"{self._deltatwin_dir}/{name}/{version}/workflow.*")
        if not path:
            raise FileNotFoundError(
                f"No workflow found for deltatwin: "
                f"{name} version {version}"
            )
        workflow = path[0]
        with open(workflow) as workflow_file:
            if workflow.endswith(".json"):
                return json.load(workflow_file)
            if workflow.endswith(".yaml") or workflow.endswith(".yml"):
                return yaml.safe_load(workflow_file)
        raise RuntimeError(f"Unsupported workflow: {workflow}")

    def get_data(self, run_id: uuid.UUID, job_id: str, basename: str) -> IO:
        path = os.path.join(self._run_dir, str(run_id), job_id, basename)
        return open(path, "rb")

    def set_data(
        self, run_id: uuid.UUID, job_id: str, basename: str, stream: IO
    ) -> str:
        path = os.path.join(self._run_dir, str(run_id), job_id, basename)
        os.makedirs(os.path.dirname(path))
        buf_size = 16 * 1024
        with open(path, "wb") as f:
            buf = stream.read(buf_size)
            while buf:
                f.write(buf)
                buf = stream.read(buf_size)
        return f"runs/{str(run_id)}/{job_id}/{basename}"

    def get_data_info(self, data_path) -> Dict:
        infos = {}
        path = os.path.join(self._base_dir, data_path)
        with open(path, "rb") as f:
            digest = hashlib.file_digest(f, "md5")
        infos["size"] = os.stat(path).st_size
        infos["checksum"] = f"md5@{digest.hexdigest()}"
        return infos

    def find_data(
        self, run_id: uuid.UUID, job_id: str, _glob: str
    ) -> List[AnyStr]:
        pattern = f"{self._run_dir}/{str(run_id)}/{job_id}/{_glob}"
        return [os.path.relpath(p, self._base_dir) for p in glob.glob(pattern)]

    def remove_run(self, run_id: uuid.UUID) -> None:
        path = os.path.join(self._run_dir, str(run_id))
        shutil.rmtree(path, ignore_errors=True)

    def push_deltatwin(self, directory, name: str, version: str) -> None:
        path = os.path.join(self._deltatwin_dir, name, version)
        self.LOGGER.info(f"pushing delta twin {name} version"
                         f" {version} into {path}")
        if not os.path.isdir(path):
            shutil.copytree(directory, path)

    def remove_deltatwin(self, name: str, version: str) -> None:
        path = os.path.join(self._deltatwin_dir, name, version)
        shutil.rmtree(path, ignore_errors=True)

    def has_deltatwin(self, name: str, version: str) -> bool:
        path = os.path.join(self._deltatwin_dir, name, version)
        return os.path.isdir(path)


class S3StorageManager(DeltaStorageManager):
    LOGGER = logging.Logger("S3StorageManager")

    def __init__(self, entry_point: str, key: str, secret: str, bucket: str):
        self._s3fs = s3fs.S3FileSystem(
            anon=False, endpoint_url=entry_point, key=key, secret=secret
        )
        self._s3 = boto3.client("s3",
                                endpoint_url=entry_point,
                                aws_access_key_id=key,
                                aws_secret_access_key=secret)
        self._bucket = bucket

    def get_full_path(self, path):
        return os.path.join("s3://", self._bucket, path)

    def get_models_filepaths(self, name, version, model_path):
        path = os.path.join("deltatwins", name, version, model_path)
        return self.get_full_path(path)

    def add_run_directory(self, run_id: uuid.UUID) -> None:
        directory = f"{self._bucket}/runs/{run_id}"
        if not self._s3fs.isdir(directory):
            self._s3fs.mkdirs(directory)

    def get_deltatwin_manifest(self, name: str, version: str) -> Manifest:
        manifest_path = f"{self._bucket}/deltatwins/{name}/{version}"
        if not self._s3fs.isdir(manifest_path):
            raise FileNotFoundError(f"Deltatwin {name} version "
                                    f"{version} not found")
        paths = self._s3fs.glob(os.path.join(manifest_path, "manifest.*"))
        with self._s3fs.open(paths[0], mode="rb") as manifest_file:
            return parse(manifest_file)

    def get_deltatwin_workflow(self, name: str, version: str) -> dict:
        glob_pattern = f"{self._bucket}/deltatwins/{name}/{version}/workflow.*"
        workflow_glob = self._s3fs.glob(glob_pattern)
        if not workflow_glob:
            raise FileNotFoundError(
                "No workflow found for deltatwin" f"{name} version {version}"
            )
        workflow = workflow_glob[0]
        with self._s3fs.open(workflow, mode="rb") as workflow_file:
            if workflow.endswith(".json"):
                return json.load(workflow_file)
            if workflow.endswith(".yml") or workflow.endswith(".yaml"):
                return yaml.safe_load(workflow_file)
        raise RuntimeError(
            "Invalid workflow file:" f"{os.path.basename(workflow)}"
        )

    def get_data(self, run_id: uuid.UUID, job_id: str, basename: str) -> IO:
        path = f"runs/{str(run_id)}/{job_id}/{basename}"
        try:
            params = {"Bucket": self._bucket, "Key": path}
            url = self._s3.generate_presigned_url(
                'get_object', Params=params, ExpiresIn=15
            )
            return urllib.request.urlopen(url)
        except ClientError as e:
            self.LOGGER.error(f'Download error: {e}')
            raise FileNotFoundError(f"No object found at: {path}")

    def set_data(
        self, run_id: uuid.UUID, job_id: str, basename: str, stream: IO
    ):
        rel_path = f"runs/{str(run_id)}/{job_id}/{basename}"
        self._s3.put_object(
            Bucket=self._bucket, Body=stream.read(), Key=rel_path
        )
        return rel_path

    def find_data(
        self, run_id: uuid.UUID, job_id: str, _glob: str
    ) -> List[AnyStr]:
        pattern = f"{self._bucket}/runs/{str(run_id)}/{job_id}/{_glob}"
        globs = self._s3fs.glob(pattern)
        return [os.path.relpath(p, start=self._bucket) for p in globs]

    def get_data_info(self, data_path) -> Dict:
        infos = {}
        response = self._s3.head_object(
            Bucket=self._bucket, Key=data_path, ChecksumMode="ENABLED"
        )
        infos["size"] = response["ContentLength"]
        # Checksum is stored in the Etag entry, surrounded by '"', hence [1:-1]
        infos["checksum"] = f"md5@{response['ETag'][1:-1]}"
        return infos

    def remove_run(self, run_id: uuid.UUID) -> None:
        directory = f"{self._bucket}/runs/{str(run_id)}"
        if self._s3fs.isdir(directory):
            self.LOGGER.info(f"Deleting run: {run_id}")
            self._s3fs.delete(directory, recursive=True)

    def push_deltatwin(self, directory, name: str, version: str) -> None:
        dest_dir = f"deltatwins/{name}/{version}"
        self.LOGGER.warning(f"Fetching deltatwin: {name} version {version}")
        for file in os.listdir(directory):
            self._s3.upload_file(
                Filename=os.path.join(directory, file),
                Bucket=self._bucket,
                Key=os.path.join(dest_dir, file),
            )

    def remove_deltatwin(self, name: str, version: str) -> None:
        directory = f"{self._bucket}/deltatwins/{name}/{version}"
        if not self._s3fs.isdir(directory):
            self.LOGGER.warning(f"Deleting deltatwin: "
                                f"{name} version {version}")
            self._s3fs.delete(directory, recursive=True)

    def has_deltatwin(self, name: str, version: str) -> bool:
        prefix = f"deltatwins/{name}/{version}"
        result = self._s3.list_objects_v2(Bucket=self._bucket, Prefix=prefix)
        return result.get('KeyCount', 0) > 0



import abc
import os
from typing import Any

from delta.manifest.parser import InputModel
from delta.run.api.model import PrimitiveParameterModel, DataParameterModel
from delta.run.utils import decrypt_with_rsa
from delta.run.config import Settings

_settings: Settings = None


def _get_private_key():
    global _settings
    if _settings is None:
        _settings = Settings()
    return _settings.private_key


class Placeholder(abc.ABC):
    @abc.abstractmethod
    def evaluate(self, secure: bool = False) -> Any:
        raise NotImplementedError


class InputModelPlaceholder(Placeholder):
    _separator: str = " "

    def __init__(self, input_model: InputModel, value: Any):
        super().__init__()
        self._input = input_model
        self._value = value
        self._factory = {
            "boolean": self.__evaluate_boolean,
            "integer": self.__evaluate_number,
            "number": self.__evaluate_number,
            "string": self.__evaluate_string,
            "secret": self.__evaluate_secret,
            "Data": self.__evaluate_data,
            "stdout": self.__evaluate_data,
        }

    def __evaluate_boolean(self) -> str:
        param: PrimitiveParameterModel | InputModel = (
                self._value or self._input
        )
        if param is not None and param.value is not None:
            value: bool | None = None
            if isinstance(param, InputModel):
                value = bool(param.value)
            if isinstance(param, PrimitiveParameterModel):
                value = param.value
            if self._input.prefix and value:
                return f"{self._input.prefix}"
        return ""

    def __evaluate_number(self) -> str:
        param: PrimitiveParameterModel | InputModel = (
                self._value or self._input
        )
        if param is not None and param.value is not None:
            if self._input.type == 'integer':
                value = int(param.value)
            else:
                value = float(param.value)
            if self._input.prefix:
                return f"{self._input.prefix}{self._separator}{value}"
            return str(value)
        return ""

    def __evaluate_string(self) -> str:
        param: PrimitiveParameterModel | InputModel = (
                self._value or self._input
        )
        if param is not None and param.value:
            if self._input.prefix:
                return f"{self._input.prefix}{self._separator}{param.value}"
            return param.value
        return ""

    def __evaluate_secret(self, secure: bool = False) -> str:
        value = self._value or self._input.value

        if value:
            decrypted_value = decrypt_with_rsa(
                value.secret_value, private_key=_get_private_key()
            )

            if secure:
                if self._input.prefix:
                    return f"{self._input.prefix}{self._separator}***"
                return "***"

            if self._input.prefix:
                return (
                    f"{self._input.prefix}"
                    f"{self._separator}"
                    f"{decrypted_value}"
                )

            return decrypted_value

        return ""

    def __evaluate_data(self) -> str:
        param: DataParameterModel | InputModel = self._value or self._input
        value: DataParameterModel | None = None
        if isinstance(param, DataParameterModel):
            value = param
        if isinstance(param, InputModel) and param.value is not None:
            value = DataParameterModel(**param.value)

        if value is not None:
            path = os.path.join('/s3', value.path)
            if self._input.prefix:
                return f"{self._input.prefix}{self._separator}{path}"
            return path

        return ""

    def evaluate(self, secure: bool = False) -> str:
        try:
            if self._input.type == 'secret':
                return self._factory[self._input.type](secure)
            return self._factory[self._input.type]()
        except KeyError:
            return ""


class JobBuilder(abc.ABC):
    @staticmethod
    def placeholder_factory(model_part: InputModel, value: Any) -> Placeholder:
        if isinstance(model_part, InputModel):
            return InputModelPlaceholder(model_part, value)
        raise NotImplementedError

    @abc.abstractmethod
    def generate_dockerfile(self) -> str:
        raise NotImplementedError

    @staticmethod
    def generate_dockerfile_ctx() -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def build_command(self):
        raise NotImplementedError



import unittest
from unittest import mock
from delta.vcs.handler import GitlabAPI


class TestUpdateMethod(unittest.TestCase):
    @mock.patch("delta.vcs.handler.gitlab_api.requests.put")
    @mock.patch("delta.vcs.handler.gitlab_api.requests.get")
    def test_with_wrong_token(self,
                              request_mock_get,
                              request_mock_put):
        request_mock_get.return_value = mock.Mock(
            status_code=200,
            json=lambda: {"id": 1,
                          "name": "starter-kit",
                          "description": "starter",
                          "title": "starter-kit"}
        )

        request_mock_put.return_value = mock.Mock(
            status_code=401,
            json=lambda: {
                "401": "Not authorized"
            }
        )
        glapi = GitlabAPI('https://git.gael.fr', 'token')

        res = glapi.update_topic(
                1, description="starter kit"
            )
        self.assertEqual(
            res.status_code, 401)

    @mock.patch('delta.vcs.handler.gitlab_api.requests.get')
    def test_with_topic_id_dont_exist(self,
                                      request_mock):
        request_mock.return_value = mock.Mock(
                        status_code=401,
                        json=lambda: {
                            "401": "Not authorized"
                        })
        glapi = GitlabAPI('https://git.gael.fr', 'token')
        self.assertIsNone(glapi.update_topic(-1, description="starter kit"))


class TestAddMethod(unittest.TestCase):
    @mock.patch("delta.vcs.handler.gitlab_api.requests.get")
    def test_with_wrong_token_and_wrong_id(self,
                                           request_mock):
        request_mock.return_value = mock.Mock(
                        status_code=401,
                        json=lambda: {
                            "401": "Not authorized"
                        })
        glapi = GitlabAPI('https://git.gael.fr', 'token')

        self.assertIsNone(glapi.add_topic_to_project(-1, "starter-kit"))

    @mock.patch("delta.vcs.handler.gitlab_api.requests.put")
    @mock.patch("delta.vcs.handler.gitlab_api.requests.get")
    def test_add_with_wrong_token(self,
                                  request_mock_get,
                                  request_mock_put):
        request_mock_get.return_value = mock.Mock(
                        status_code=200,
                        json=lambda: {"id": 1,
                                      "topics": ['test']})

        request_mock_put.return_value = mock.Mock(
                        status_code=401,
                        json=lambda: {
                            "401": "Not authorized"
                        })

        glapi = GitlabAPI('https://git.gael.fr', 'token')

        self.assertEqual(
            glapi.add_topic_to_project(
                1,
                "starter-kit"
            ).status_code, 401)


class TestGetMethods(unittest.TestCase):
    @mock.patch("delta.vcs.handler.gitlab_api.requests.get")
    def test_get_topic_by_name_not_in_list(self,
                                           request_mock):
        request_mock.return_value = mock.Mock(status_code=200,
                                              json=lambda: [
                                                    {
                                                        "id": 1,
                                                        "name": "test"
                                                    },
                                                    {
                                                        "id": 2,
                                                        "name": "toto"
                                                    }])

        glapi = GitlabAPI('https://git.gael.fr', 'token')
        self.assertIsNone(glapi.get_topic_by_name('toti'))

    @mock.patch("delta.vcs.handler.gitlab_api.requests.get")
    def test_get_topic_by_name_in_list(self,
                                       request_mock):
        request_mock.return_value = mock.Mock(status_code=200,
                                              json=lambda: [
                                                    {
                                                        "id": 1,
                                                        "name": "test"
                                                    },
                                                    {
                                                        "id": 2,
                                                        "name": "toto"
                                                    }])

        glapi = GitlabAPI('https://git.gael.fr', 'token')
        self.assertIsNotNone(glapi.get_topic_by_name('toto'))
        self.assertEqual(
            glapi.get_topic_by_name('toto'),
            {"id": 2, "name": "toto"}
        )

    @mock.patch('delta.vcs.handler.gitlab_api.requests.get')
    def test_get_topics_with_wrong_token(self,
                                         request_mock):
        request_mock.return_value = mock.Mock(status_code=200,
                                              json=lambda: {
                                                  "topics": [
                                                    {
                                                        "id": 1,
                                                        "name": "test"
                                                    },
                                                    {
                                                        "id": 2,
                                                        "name": "toto"
                                                    }]})

        glapi = GitlabAPI('https://git.gael.fr', 'token')
        self.assertEqual(glapi.get_topics().status_code, 200)


class TestCreateMethod(unittest.TestCase):
    @mock.patch('delta.vcs.handler.gitlab_api.requests.post')
    @mock.patch('delta.vcs.handler.gitlab_api.requests.get')
    def test_create_topic_with_wrong_token(self,
                                           request_mock_get,
                                           request_mock_post):
        request_mock_get.return_value = mock.Mock(status_code=401,
                                                  json=lambda: {
                                                      "401": "Not authorized"
                                                    })
        request_mock_post.return_value = mock.Mock(
                        status_code=401,
                        json=lambda: {
                            "401": "Not authorized"
                        })
        glapi = GitlabAPI('https://git.gael.fr', 'token')
        self.assertEqual(glapi.create_topic('hurricane',
                                            'hurricane'
                                            ).status_code, 401)



from delta.store.interfaces import IData
from delta.store.interfaces import IDataIO
import drb.core.signature
from drb.topics import resolver
import io


class DrbDataIOImpl(IDataIO):
    def __init__(self, src: IData, dst=None):
        super().__init__(src, dst)

    def _read(self):
        path = self._src.info['path']
        drb_node = resolver.create(path)
        if drb_node.has_impl(io.BufferedIOBase):
            return drb_node.get_impl(io.BufferedIOBase)
        if drb_node.has_impl(io.BytesIO):
            return drb_node.get_impl(io.BytesIO)
        raise TypeError("DRB Impl not supported...")

    def _write(self, data):
        self._dst.write(data.read())



import datetime as dt
import enum
from typing import List

from sqlalchemy import Enum, ForeignKey
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship


class RunStatus(enum.Enum):
    CREATED = "created"
    RUNNING = "running"
    SUCCESS = "success"
    CANCELLED = "cancelled"
    ERROR = "error"


class ParameterKind(enum.Enum):
    INPUT = "INPUT"
    OUTPUT = "OUTPUT"


class Base(DeclarativeBase):
    pass


class Parameter(Base):
    __tablename__ = 'parameters'
    id: Mapped[int] = mapped_column(primary_key=True)
    run_id: Mapped[str] = mapped_column(ForeignKey('runs.id'))
    name: Mapped[str]
    param_type: Mapped[str]
    type: Mapped[str]
    kind: Mapped[ParameterKind]

    __mapper_args__ = {
        'polymorphic_on': 'param_type',
        'polymorphic_identity': 'parameter'
    }


class PrimitiveParameter(Parameter):
    value: Mapped[str] = mapped_column(nullable=True)

    __mapper_args__ = {'polymorphic_identity': 'primitive_parameter'}


class SecretParameter(Parameter):
    secret_value: Mapped[bytes] = mapped_column(nullable=True)

    __mapper_args__ = {'polymorphic_identity': 'secret_parameter'}


class DataParameter(Parameter):
    url: Mapped[str] = mapped_column(nullable=True)
    path: Mapped[str] = mapped_column(nullable=True)
    size: Mapped[int] = mapped_column(nullable=True)
    checksum: Mapped[str] = mapped_column(nullable=True)

    __mapper_args__ = {'polymorphic_identity': 'data_parameter'}


class RunContext(Base):
    __tablename__ = 'runs'

    id: Mapped[str] = mapped_column(primary_key=True)
    deltatwin_id: Mapped[str]
    deltatwin_version: Mapped[str] = mapped_column(default="main")
    owner: Mapped[str]
    date_created: Mapped[dt.datetime] = mapped_column(
        default=dt.datetime.now(dt.timezone.utc))
    status: Mapped[RunStatus] = mapped_column(Enum(RunStatus))
    return_code: Mapped[int] = mapped_column(default=0)
    message: Mapped[str] = mapped_column(default="")
    parent_run: Mapped[str] = mapped_column(default="")
    inputs: Mapped[List[Parameter]] = relationship(
        "Parameter",
        primaryjoin="and_(RunContext.id==Parameter.run_id,"
                    f"Parameter.kind=='{ParameterKind.INPUT.value}')",
        cascade="all, delete-orphan"
    )
    outputs: Mapped[List[Parameter]] = relationship(
        "Parameter",
        primaryjoin="and_(RunContext.id==Parameter.run_id,"
                    f"Parameter.kind=='{ParameterKind.OUTPUT.value}')",
        cascade="all, delete-orphan",
        overlaps='inputs'
    )



from __future__ import annotations

from datetime import datetime, timezone
from functools import partial
from typing import List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field

from delta.run.db.orm import (
    DataParameter,
    Parameter,
    SecretParameter,
    ParameterKind,
    PrimitiveParameter,
    RunContext,
    RunStatus,
)


class ParameterModel(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    name: str
    type: Optional[str] = None


class PrimitiveParameterModel(ParameterModel):
    value: Union[bool, int, float, str]


class SecretParameterModel(ParameterModel):
    secret_value: Union[str, bytes] = Field(repr=False, exclude=True)


class DataParameterModel(ParameterModel):
    url: str
    path: Optional[str] = Field(default=None)
    size: Optional[int] = Field(default=None)
    checksum: Optional[str] = Field(default=None)


def orm_param_to_model_param(param: Parameter) -> ParameterModel:
    if isinstance(param, SecretParameter):
        value = param.secret_value
        return SecretParameterModel(
            name=param.name, secret_value=value, type=param.type
        )
    if isinstance(param, PrimitiveParameter):
        return PrimitiveParameterModel(
            name=param.name, value=param.value, type=param.type
        )
    elif isinstance(param, DataParameter):
        return DataParameterModel(
            name=param.name,
            url=param.url,
            path=param.path,
            size=param.size,
            type=param.type,
            checksum=param.checksum,
        )
    else:
        raise ValueError(f"Unsupported parameter type: {param.type}")


def retrieve_primitive_type(value: Union[bool, int, float, str]):
    if isinstance(value, bool):
        return "boolean"
    elif isinstance(value, int):
        return "integer"
    elif isinstance(value, float):
        return "number"
    elif isinstance(value, str):
        return "string"
    raise ValueError(f"Unsupported primitive parameter type: {type(value)}")


def model_param_to_orm_param(
    param: ParameterModel, kind: ParameterKind
) -> Parameter:
    if isinstance(param, SecretParameterModel):
        return SecretParameter(
            name=param.name,
            param_type="secret_parameter",
            type="secret",
            kind=kind,
            secret_value=param.secret_value
        )
    if isinstance(param, PrimitiveParameterModel):
        return PrimitiveParameter(
            name=param.name,
            param_type="primitive_parameter",
            type=retrieve_primitive_type(param.value),
            kind=kind,
            value=param.value
        )
    elif isinstance(param, DataParameterModel):
        return DataParameter(
            name=param.name,
            param_type="data_parameter",
            type="Data",
            kind=kind,
            url=param.url,
            path=param.path,
            size=param.size,
            checksum=param.checksum,
        )


class RunContextModel(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    id: str
    deltatwin_id: str
    deltatwin_version: str = Field(default="main")
    owner: str
    date_created: datetime
    status: RunStatus = Field(default=RunStatus.CREATED)
    return_code: int = Field(default=0)
    message: str = Field(default="")
    parent_run: str = Field(default="")
    inputs: List[Union[PrimitiveParameterModel,
                       DataParameterModel, SecretParameterModel]] = Field(
        default_factory=list
    )
    outputs: List[Union[PrimitiveParameterModel,
                        DataParameterModel, SecretParameterModel]] = Field(
        default_factory=list
    )

    @classmethod
    def from_run_context(cls, run_context: RunContext) -> RunContextModel:
        return cls(
            id=run_context.id,
            deltatwin_id=run_context.deltatwin_id,
            deltatwin_version=run_context.deltatwin_version,
            owner=run_context.owner,
            status=run_context.status,
            date_created=run_context.date_created,
            message=run_context.message,
            return_code=run_context.return_code,
            parent_run=run_context.parent_run,
            inputs=[orm_param_to_model_param(p) for p in run_context.inputs],
            outputs=[orm_param_to_model_param(p) for p in run_context.outputs],
        )


class RunContextInsertModel(BaseModel):
    deltatwin_id: str
    owner: str
    date_created: datetime = Field(
        default_factory=partial(datetime.now, timezone.utc)
    )
    deltatwin_version: str = Field(default="main")
    inputs: List[Union[PrimitiveParameterModel, DataParameterModel,
                       SecretParameterModel]]


class RunContextUpdateModel(BaseModel):
    status: Optional[RunStatus] = None
    return_code: Optional[int] = None
    message: Optional[str] = None
    inputs: Optional[
        List[Union[PrimitiveParameterModel, DataParameterModel,
                   SecretParameterModel]]
    ] = None
    outputs: Optional[
        List[Union[PrimitiveParameterModel, DataParameterModel,
                   SecretParameterModel]]
    ] = None



from abc import ABC, abstractmethod
from typing import TypeVar

from .i_data_io import IDataIO

T = TypeVar('T')


class IFetcher(ABC):

    def __init__(self, dataIO: IDataIO):
        self._dataIO = dataIO

    @abstractmethod
    def fetch(self) -> T:
        raise NotImplementedError()



import unittest

from delta.vcs.handler.delta_git_handler import DeltaGitHandler


class TestClose(unittest.TestCase):
    def test_close(self):
        gh = DeltaGitHandler()
        gh._path = 'test'
        gh._initialize_flag = True
        gh.close()
        self.assertEqual(gh._path, None)
        self.assertEqual(gh._initialize_flag, False)



import unittest
import os

from delta.vcs.handler.delta_git_handler import DeltaGitHandler, DeltaException


class TestCloneRepo(unittest.TestCase):

    def test_clone_not_https_or_ssh_url(self):
        gh = DeltaGitHandler()
        with self.assertRaises(DeltaException):
            gh.clone('toto', '.')
        self.assertTrue(gh._repo is None)
        self.assertTrue(gh._path is None)
        self.assertTrue(gh._initialize_flag is False)

    def test_clone_https_url_but_not_a_repo(self):
        gh = DeltaGitHandler()
        with self.assertRaises(DeltaException):
            gh.clone('https://git.gael.fr/', '.')
        self.assertTrue(gh._repo is None)
        self.assertTrue(gh._path is None)
        self.assertTrue(gh._initialize_flag is False)

    def test_clone_https(self):
        gh = DeltaGitHandler()
        self.assertIsNone(
            gh.clone('https://gitlab.com/drb-python/drb.git', '.'))
        self.assertIsNotNone(gh._repo)
        self.assertEqual(gh._path, os.path.join(os.path.abspath('.'), 'drb'))
        self.assertTrue(gh._initialize_flag is True)



import abc
import dataclasses
import importlib
import inspect
import json
from re import S
from typing import IO, Any, Dict, List, Union

import yaml

from delta.manifest.constants import manifest_version
from delta.manifest.manifest import check_manifest, read_manifest


@dataclasses.dataclass
class Copyright:
    company: str
    years: List[int] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class License:
    name: str
    url: str
    description: str = dataclasses.field(default=None)
    copyrights: List[Copyright] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class Parameter:
    name: str
    type: str
    description: str = dataclasses.field(default=None)


@dataclasses.dataclass
class Resource(Parameter):
    value: Any = dataclasses.field(default=None)


@dataclasses.dataclass
class Input(Parameter):
    value: Any = dataclasses.field(default=None)


@dataclasses.dataclass
class InputModel(Input):
    prefix: str = dataclasses.field(default=None)


@dataclasses.dataclass
class Output(Parameter):
    pass


@dataclasses.dataclass
class OutputModel(Output):
    glob: str = dataclasses.field(default=None)


@dataclasses.dataclass
class Model:
    path: str
    type: str
    parameters: Dict[str, Any] = dataclasses.field(default_factory=dict)
    inputs: Dict[str, InputModel] = dataclasses.field(default_factory=dict)
    outputs: Dict[str, OutputModel] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class Dependency:
    id: str
    version: str


@dataclasses.dataclass
class Manifest:
    name: str
    description: str
    license: License
    short_description: str = dataclasses.field(default_factory=str)
    owner: str = dataclasses.field(default_factory=str)
    resources: Dict[str, Resource] = \
        dataclasses.field(default_factory=dict)
    inputs: Dict[str, Input] = dataclasses.field(default_factory=dict)
    outputs: Dict[str, Output] = dataclasses.field(default_factory=dict)
    models: Dict[str, Model] = dataclasses.field(default_factory=dict)
    dependencies: Dict[str, Dependency] = dataclasses.field(
        default_factory=dict
    )


class ManifestParser(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def parse_license(cls, data: dict) -> License:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def parse_resource(cls, data: dict) -> Resource:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def parse_input(cls, data: dict) -> Input:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def parse_output(cls, data: dict) -> Output:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def parse_model(cls, data: dict) -> Model:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def parse(cls, data: dict) -> Manifest:
        raise NotImplementedError


def _manifest_parser_filter(obj: Any) -> bool:
    return (
        inspect.isclass(obj)
        and issubclass(obj, ManifestParser)
        and obj != ManifestParser
    )


def _parse_from_dict(data: dict) -> Manifest:
    version = data.get("deltaVersion", manifest_version)
    validated = check_manifest(data, version=version, verbose=True)
    if validated is False:
        raise ValueError("Invalid manifest content")

    try:
        module = importlib.import_module(
            f"delta.manifest.v{version.replace('.', '_')}"
        )
    except ModuleNotFoundError:
        module = importlib.import_module(
            f"delta.manifest.v{manifest_version.replace('.', '_')}"
        )
    members = inspect.getmembers(module, predicate=_manifest_parser_filter)

    if not members:
        raise RuntimeError(
            f"No DeltaTwin manifest parser found for version: {version}"
        )
    parser: ManifestParser = members[-1][1]
    return parser.parse(data)


def _parse_from_path(path: str) -> Manifest:
    data = read_manifest(path)
    return _parse_from_dict(data)


def _parse_from_stream(stream: IO) -> Manifest:
    try:
        data = yaml.safe_load(stream)
    except yaml.YAMLError:
        data = json.load(stream)
    return _parse_from_dict(data)


def parse(obj: Union[str, dict, IO]) -> Manifest:
    if isinstance(obj, str):
        return _parse_from_path(obj)
    elif isinstance(obj, IO) or hasattr(obj, "read"):
        return _parse_from_stream(obj)
    return _parse_from_dict(obj)



import unittest
import os

from delta.vcs.handler.delta_git_handler import DeltaGitHandler, DeltaException


class TestWithOpen(unittest.TestCase):

    def test_with_open(self):
        with DeltaGitHandler() as repo:
            repo.create('test_with_open')
            self.assertIsNotNone(repo._repo)
            path = os.path.join(
                os.path.abspath('.'),
                'test_with_open')
            self.assertTrue(repo._path == path)
            self.assertTrue(repo._initialize_flag)



import abc


class IDrive(abc.ABC):

    def __init__(self):
        pass

    def clone(self, url: str, path_to_save: str = '.'):
        raise NotImplementedError()

    def init(self, path_to_save: str):
        raise NotImplementedError()

    def status(self):
        raise NotImplementedError()

    def tag(self, **kwargs):
        raise NotImplementedError()

    def commit(self, message: str):
        raise NotImplementedError()

    def pull(self, origin_name: str):
        raise NotImplementedError()

    def push(self, **kwargs):
        raise NotImplementedError()

    def branch(self, **kwargs):
        raise NotImplementedError()

    def checkout(self, branch_name: str):
        raise NotImplementedError()

    def reset(self):
        raise NotImplementedError()

    def add_dependency(self):
        raise NotImplementedError()

    def add_resource(self, **kwargs):
        raise NotImplementedError()

    def check(self):
        raise NotImplementedError()

    def fetch(self, origin_name: str):
        raise NotImplementedError()

    def sync(self):
        raise NotImplementedError()



from git import Repo
from typing import Any
from urllib.parse import urlparse
from delta.manifest import manifest
import os
import re
import shutil
import logging

TEMPLATE_DIR = os.path.join(
                            os.path.dirname(os.path.abspath(__file__)),
                            'template')

FILES_TO_ADD = ['.delta',
                'artifacts',
                'models',
                'resources',
                'sources',
                '.gitignore',
                'manifest.json']

logger = logging.getLogger('Delta')


class DeltaException(Exception):
    pass


def url_validator(url):
    ssh_pattern = (r"(ssh:\/\/)?(?:[a-z_][a-z0-9_]{0,30})@"
                   r"((?:[a-z0-9-_]+\.)+[a-z0-9]+)(?::([0-9]{0,5}))?"
                   r"(\/[^\0\n]+)?")
    return urlparse(url).scheme == 'https' or bool(re.match(ssh_pattern, url))


def get_repo_name_from_url(path):
    return path.split('.git')[0].split('/')[-1]


class DeltaGitHandler:

    def __init__(self):
        self._repo = None
        self._initialize_flag = False
        self._path = None

    def create(self, path_to_save: str):
        try:
            _path = os.path.abspath(path_to_save)
            shutil.copytree(TEMPLATE_DIR, _path)
            manifest.write_manifest(os.path.join(_path, 'manifest.json'))
            _repo = Repo.init(_path)
            if self._initialize_flag:
                self.close()
            self._repo = _repo
            self._path = _path
            self._initialize_flag = True
            self.add(FILES_TO_ADD, force=True)
            self.commit('Initialize Twin')
            logger.info(f"Twin has been initialized at {_path}")
        except Exception:
            raise DeltaException("Error when creating the repo")

    def set_config(self, username, email):
        try:
            self._repo.config_writer().set_value("user",
                                                 "name",
                                                 username).release()
            self._repo.config_writer().set_value("user",
                                                 "email",
                                                 email).release()
            logger.info(f"Config :\n\tusername = {username}\n\t{email}")
        except Exception:
            raise DeltaException("Error when setting config")

    def load_local_repo(self, path: str):
        try:
            _path = os.path.abspath(path)
            _repo = Repo(_path)
            if self._initialize_flag:
                self.close()
            self._repo = _repo
            self._path = _path
            self._initialize_flag = True
            logger.debug(f"Repo {path} has been loaded")
        except Exception:
            raise DeltaException("Error when loading existing repo...")

    def clone(self, path_from: str, path_to_save: str):
        try:
            if not url_validator(path_from):
                raise DeltaException(f'Wrong URL : {path_from}... Verify it.')
            else:
                path_with_repo_name = os.path.join(
                                                path_to_save,
                                                get_repo_name_from_url(
                                                    path_from
                                                )
                                            )
                _repo = Repo.clone_from(path_from, path_with_repo_name)
                _path = os.path.abspath(path_with_repo_name)
                logger.info(f"Repo {path_from} has been cloned "
                            f"and store at {_path}")
                if self._initialize_flag:
                    self.close()
                self._repo = _repo
                self._path = _path
                self._initialize_flag = True
        except Exception:
            raise DeltaException("Error when cloning project. \n"
                                 f"Maybe the path {path_from} doesn't exist. "
                                 "Or check the path where you want to save "
                                 "the repo.")

    def commit(self, message: str):
        if self._initialize_flag:
            try:
                commit = self._repo.index.commit(message)
                logger.info(f"New commit : {commit.name_rev}")
            except Exception:
                raise DeltaException("Error when commiting.")
        else:
            raise DeltaException("Please initialize your repository "
                                 "in order to commit.")

    def add(self, files_to_add: list, force: bool = False):
        if self._initialize_flag:
            try:
                files = []
                if force:
                    files = files_to_add
                else:
                    files = filter(
                        lambda files: not files.startswith('resources/'),
                        files_to_add)
                res = self._repo.index.add(files)
                logger.debug(f"{files_to_add} has been added")
                return res
            except Exception:
                raise DeltaException("Error when adding files "
                                     f"{files_to_add}.")
        else:
            raise DeltaException("Please initialize your repository "
                                 "in order to add files.")

    def pull(self, remote_name: str):
        if self._initialize_flag:
            try:
                remote = None
                if remote_name == '' or remote_name is None:
                    remote = self._repo.remotes.origin
                else:
                    remote = self._repo.remote(name=remote_name)
                if remote is not None:
                    remote.pull(refspec=f"{self._repo.active_branch}:"
                                f"{self._repo.active_branch}")
            except Exception:
                raise DeltaException("Error when pulling.")
        else:
            raise DeltaException("Please initialize your repository "
                                 "in order to pull")

    def push(self, remote_name: str = None, ref_name: str = None):
        if self._initialize_flag:
            try:
                if remote_name == '' or remote_name is None:
                    remote = self._repo.remotes.origin
                else:
                    remote = self._repo.remote(name=remote_name)

                if ref_name == '' or ref_name is None:
                    ref_to_push = self._repo.active_branch
                else:
                    ref_to_push = ref_name
                remote.push(
                            refspec=f"{ref_to_push}:"
                            f"{ref_to_push}"
                        )
                logger.info("Pushed on remote branch "
                            f"{remote_name}/{ref_to_push}")
            except Exception:
                try:
                    logger.warning("Will try with --set-upstream...")
                    self._repo.git.push('--set-upstream',
                                        remote.name,
                                        ref_to_push)
                except Exception:
                    raise DeltaException("Error when trying to push.")
        else:
            raise DeltaException("Please initialize your repository "
                                 "in order to push")

    def fetch(self, remote_name: str):
        if self._initialize_flag:
            try:
                if remote_name == '' or remote_name is None:
                    remote = self._repo.remotes.origin
                else:
                    remote = self._repo.remote(name=remote_name)
                remote.fetch()
            except Exception:
                raise DeltaException("Error when call fetch.")
        else:
            raise DeltaException("Please initialize your repository "
                                 "in order to fetch")

    def create_remote(self, remote_name: str, url: str):
        if self._initialize_flag:
            try:
                self._repo.create_remote(remote_name, url)
                logger.info(f"Remote {remote_name} at {url} "
                            "has been created")
            except Exception:
                raise DeltaException("Error when creating remote.")
        else:
            raise DeltaException("Please initialize your repository "
                                 "in order to create a remote")

    def list_remotes(self):
        if self._initialize_flag:
            try:
                remotes = [remote.name for remote in self._repo.remotes]
                return remotes
            except Exception:
                raise DeltaException("Error when executing command "
                                     "'git remote show'")
        else:
            raise DeltaException("Please initialize your repository "
                                 "in order to list remotes")

    def status(self):
        if self._initialize_flag:
            try:
                return self._repo.head.reference.log()
            except Exception:
                raise DeltaException("Error when get status.")
        else:
            raise DeltaException("Please initialize your repository "
                                 "in order to get repository status")

    def checkout(self, branch_name: str):
        if self._initialize_flag:
            try:
                self._repo.git.checkout(branch_name)
                logger.info(f"Checkout to branch : {branch_name}")
            except Exception:
                try:
                    self._repo.git.checkout('-b', branch_name)
                    logger.info(f"Checkout to branch : {branch_name}")
                except Exception:
                    raise DeltaException("Error when checkout to "
                                         f"branch {branch_name}. "
                                         "Please verify the name "
                                         "of your branch.")
        else:
            raise DeltaException("Please initialize your repository "
                                 "in order to checkout.")

    def branch(self, branch_name: str):
        if self._initialize_flag:
            try:
                self._repo.git.branch(branch_name)
                logger.info(f"{branch_name} has been created.")
            except Exception:
                raise DeltaException("Error when executing command "
                                     "'git branch <branch name'.")
        else:
            raise DeltaException("Please initialize your repository "
                                 "in order to create a new branch.")

    def remove_branch(self, branch_name: str):
        if self._initialize_flag:
            try:
                self._repo.git.branch('-D', branch_name)
                logger.info(f"{branch_name} has been removed.")
            except Exception:
                raise DeltaException("Error when executing command "
                                     "'git branch -D <branch name'.")
        else:
            raise DeltaException("Please initialize your repository "
                                 "in order to create a new branch.")

    def list_branches(self):
        if self._initialize_flag:
            branches = [branch.name for branch in self._repo.branches]
            remotes = self._repo.remotes
            for remote in remotes:
                for branch_ref in remote.refs:
                    branches.append(branch_ref.name)
            return branches

        else:
            raise DeltaException("Please initialize your repository "
                                 "in order to list all branches.")

    def tag(self, tag_name: str, message: str = ''):
        if self._initialize_flag:
            try:
                res = self._repo.create_tag(tag_name, message=message)
                logger.info(f"New tag {tag_name} has been created")
                return str(res)
            except Exception:
                raise DeltaException("Error when trying to create "
                                     f"new tag {tag_name}.")
        else:
            raise DeltaException("Please initialize your repository "
                                 "in order to tag")

    def list_tag(self) -> list:
        if self._initialize_flag:
            try:
                res = [tag.name for tag in self._repo.tags]
                return res
            except Exception:
                raise DeltaException("Error when trying to listing tags. ")
        else:
            raise DeltaException("Please initialize your repository "
                                 "in order to list tags")

    def remove_tag(self, tag_name: str):
        if self._initialize_flag:
            try:
                self._repo.git.tag('-d', tag_name)
                logger.info(f"{tag_name} has been removed.")
            except Exception:
                raise DeltaException("Error when trying to remove "
                                     f"tag {tag_name}.")
        else:
            raise DeltaException("Please initialize your repository "
                                 "in order to tag")

    def close(self):
        if self._initialize_flag:
            if self._repo is not None:
                self._repo.close()
                logger.debug("Repo has been closed...")
            self._path = None
            self._initialize_flag = False
            logger.debug("Everything has been closed correctly...")

    def __enter__(self) -> "DeltaGitHandler":
        return self

    def __exit__(self, *args: Any) -> None:
        self.close()

    def __del__(self) -> None:
        try:
            self.close()
        except Exception:
            pass



import unittest
import os

from delta.core import DeltaCore
from delta.vcs.handler import DeltaGitHandler


class TestCoreDrive(unittest.TestCase):
    def test_drive_clone(self):
        with DeltaCore() as core:
            try:
                core.drive_clone('toto')
            except Exception:
                pass

    def test_core_drive_init(self):
        with DeltaCore() as core:
            core.drive_init('test_core_drive_init')
            self.assertTrue(os.path.exists('test_core_drive_init'))

    def test_core_drive_status(self):
        with DeltaCore() as core:
            core.drive_status()

    def test_core_drive_tag(self):
        with DeltaCore() as core:
            core.drive_init('test_core_drive_tag')
            core._drive.repo_directory = 'test_core_drive_tag'
            self.assertEqual([], core.drive_tag(list_tags=True))
            core.drive_tag(tag_name='v1.0.0')
            self.assertEqual(['v1.0.0'], core.drive_tag(list_tags=True))
            core.drive_tag(tag_name='v1.0.0',
                           delete=True,
                           message='Hello World!')
            self.assertEqual([], core.drive_tag(list_tags=True))

    def test_core_drive_commit(self):
        with DeltaCore() as core:
            core.drive_init('test_core_drive_commit')
            core._drive.repo_directory = 'test_core_drive_commit'
            core.drive_commit("Hello World!")
            with DeltaGitHandler() as repo:
                repo.load_local_repo('test_core_drive_commit')
                self.assertEqual(repo._repo.head.commit.message,
                                 "Hello World!")

    def test_core_drive_pull(self):
        with DeltaCore() as core:
            try:
                core.drive_init('test_core_drive_pull')
                core._drive.repo_directory = 'test_core_drive_pull'
                core.drive_pull()
            except Exception:
                pass

    def test_core_drive_push(self):
        with DeltaCore() as core:
            try:
                core.drive_init('test_core_drive_push')
                core._drive.repo_directory = 'test_core_drive_push'
                core.drive_push()
            except Exception:
                pass

    def test_core_drive_branch(self):
        with DeltaCore() as core:
            core.drive_init('test_core_drive_branch')
            core._drive.repo_directory = 'test_core_drive_branch'
            branches = core.drive_branch()
            self.assertEqual(['master'], branches)
            core.drive_branch(branch_name='toto')
            branches = core.drive_branch()
            self.assertEqual(['master', 'toto'], branches)
            core.drive_branch(branch_name='toto', delete=True)
            branches = core.drive_branch()
            self.assertEqual(['master'], branches)

    def test_core_drive_checkout(self):
        with DeltaCore() as core:
            core.drive_init('test_core_drive_checkout')
            core._drive.repo_directory = 'test_core_drive_checkout'
            branches = core.drive_branch()
            self.assertEqual(['master'], branches)
            core.drive_checkout('toto')
            branches = core.drive_branch()
            self.assertEqual(['master', 'toto'], branches)
            with self.assertRaises(Exception):
                core.drive_branch(branch_name='toto', delete=True)
            core.drive_checkout('master')
            core.drive_branch(branch_name='toto', delete=True)
            branches = core.drive_branch()
            self.assertEqual(['master'], branches)

    def test_core_drive_reset(self):
        with DeltaCore() as core:
            core.drive_init('test_core_drive_reset')
            core._drive.repo_directory = 'test_core_drive_reset'
            core.drive_reset()

    def test_core_drive_add_dependency(self):
        with DeltaCore() as core:
            core.drive_init('test_core_drive_add_dependency')
            core._drive.repo_directory = 'test_core_drive_add_dependency'
            core.drive_add_dependency()

    def test_core_drive_add_resource(self):
        with DeltaCore() as core:
            try:
                core.drive_init('test_core_drive_add_resource')
                core._drive.repo_directory = 'test_core_drive_add_resource'
                core.drive_add_resource(name="toto", path="goodurl/toto")
            except Exception:
                pass

    def test_core_drive_check(self):
        with DeltaCore() as core:
            core.drive_init('test_core_drive_check')
            core._drive.repo_directory = 'test_core_drive_check'
            self.assertTrue(core.drive_check())

    def test_core_drive_fetch(self):
        with DeltaCore() as core:
            try:
                core.drive_init('test_core_drive_fetch')
                core._drive.repo_directory = 'test_core_drive_fetch'
                core.drive_fetch()
            except Exception:
                pass

    def test_core_drive_sync(self):
        with DeltaCore() as core:
            core.drive_init('test_core_drive_sync')
            core._drive.repo_directory = 'test_core_drive_sync'
            core.drive_sync()



import unittest

from delta.store.impl import StoreImpl
from delta.manifest import manifest
import os

RESOURCE_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            'resources')


class TestStore(unittest.TestCase):
    def test_add_resource(self):
        store = StoreImpl()
        store.add_resource(resources_path=RESOURCE_DIR,
                           name="drb.zip",
                           path="https://gitlab.com/drb-python/drb/"
                           "-/archive/main/drb-main.zip")
        _manifest = manifest.read_manifest()
        self.assertEqual(_manifest['resources']['items'][0]["name"], 'drb.zip')
        self.assertEqual(
            _manifest['resources']['items'][0]["source_url"],
            "https://gitlab.com/drb-python/drb/""-/archive/main/drb-main.zip")
        self.assertFalse(os.path.exists(os.path.join(
            RESOURCE_DIR, "drb.zip"
        )))
        self.assertEqual(store.get_resources()[0]['name'], 'drb.zip')
        store.remove_resource(name='drb.zip')
        self.assertEqual(len(store.get_resources()), 0)
        store.add_resource(resources_path=RESOURCE_DIR,
                           name="drb.zip",
                           path="https://gitlab.com/drb-python/drb/"
                           "-/archive/main/drb-main.zip",
                           download=True)
        _manifest = manifest.read_manifest()
        self.assertEqual(_manifest['resources']['items'][0]["name"], 'drb.zip')
        self.assertEqual(
            _manifest['resources']['items'][0]["source_url"],
            "https://gitlab.com/drb-python/drb/""-/archive/main/drb-main.zip")
        self.assertTrue(os.path.exists(os.path.join(
            RESOURCE_DIR, "drb.zip"
        )))



import os
import unittest

from delta.manifest.parser import Model, InputModel
from delta.run.api.model import PrimitiveParameterModel, DataParameterModel
from delta.run.job.builder.python_builder import PythonRunner, SoftwarePackage


class TestPythonJobBuilder(unittest.TestCase):
    _model: Model = None

    @classmethod
    def setUpClass(cls):
        cls._model = Model(
            path="models/fake_model",
            type="python",
            parameters={},
            inputs={}
        )

    def test_build_command_bool(self):
        self._model.inputs = {
            "flag1": InputModel(
                name="flag1", type="boolean", value=True, prefix='-f1'),
            "flag2": InputModel(
                name="flag2", type="boolean", value=False, prefix='-f2'),
            "flag3": InputModel(
                name="flag3", type="boolean", prefix='-f3'),
        }
        self._model.parameters["command"] = (
            "test $(inputs.flag1) $(inputs.flag2) $(inputs.flag3)"
        )
        builder = PythonRunner(self._model)

        # not input provided
        params = {"inputs": {}}
        actual = builder.build_command(**params)
        self.assertEqual("test -f1", actual)

        # input flag1 set to False
        params["inputs"]["flag1"] = PrimitiveParameterModel(
            name="flag1", type="boolean", value=False
        )
        actual = builder.build_command(**params)
        self.assertEqual("test", actual)
        del params["inputs"]["flag1"]

        # intput flag2 set to True
        params["inputs"]["flag2"] = PrimitiveParameterModel(
            name="flag2", type="boolean", value=True
        )
        actual = builder.build_command(**params)
        self.assertEqual("test -f1 -f2", actual)
        del params["inputs"]["flag2"]

        # intput flag3 set to False
        params["inputs"]["flag3"] = PrimitiveParameterModel(
            name="flag2", type="boolean", value=False
        )
        actual = builder.build_command(**params)
        self.assertEqual("test -f1", actual)
        del params["inputs"]["flag3"]

        # intput flag3 set to True
        params["inputs"]["flag3"] = PrimitiveParameterModel(
            name="flag2", type="boolean", value=True
        )
        actual = builder.build_command(**params)
        self.assertEqual("test -f1 -f3", actual)
        del params["inputs"]["flag3"]

    def test_build_command_int(self):
        self._model.inputs = {
            "n": InputModel(name="n", type="integer", value=3, prefix='-n'),
            "i": InputModel(name="i", type="integer", prefix='-i'),
            "x": InputModel(name="x", type="integer"),
        }
        self._model.parameters["command"] = (
            "test $(inputs.n) $(inputs.i) $(inputs.x)"
        )
        builder = PythonRunner(self._model)

        params = {"inputs": {}}
        expected = "test -n 3"
        actual = builder.build_command(**params)
        self.assertEqual(expected, actual)

        params = {"inputs": {
            "n": PrimitiveParameterModel(name="n", type="integer", value=42)
        }}
        expected = "test -n 42"
        actual = builder.build_command(**params)
        self.assertEqual(expected, actual)

        params = {"inputs": {
            "n": PrimitiveParameterModel(name="n", type="integer", value=5),
            "i": PrimitiveParameterModel(name="i", type="integer", value=7)
        }}
        expected = "test -n 5 -i 7"
        actual = builder.build_command(**params)
        self.assertEqual(expected, actual)

        params = {"inputs": {
            "x": PrimitiveParameterModel(name="x", type="integer", value=31),
        }}
        expected = "test -n 3 31"
        actual = builder.build_command(**params)
        self.assertEqual(expected, actual)

    def test_build_command_string(self):
        self._model.inputs = {
            "t": InputModel(name="a", type="string", value="tab", prefix='-t'),
            "o": InputModel(name="b", type="string", prefix='-o'),
            "f": InputModel(name="b", type="string"),
        }
        self._model.parameters["command"] = (
            "test $(inputs.t) $(inputs.o) $(inputs.f)"
        )
        builder = PythonRunner(self._model)

        params = {"inputs": {}}
        expected = "test -t tab"
        actual = builder.build_command(**params)
        self.assertEqual(expected, actual)

        params = {"inputs": {
            "t": PrimitiveParameterModel(
                name="t", type="string", value="space"
            ),
        }}
        expected = "test -t space"
        actual = builder.build_command(**params)
        self.assertEqual(expected, actual)

        params = {"inputs": {
            "o": PrimitiveParameterModel(
                name="o", type="string", value="stdout"
            ),
        }}
        expected = "test -t tab -o stdout"
        actual = builder.build_command(**params)
        self.assertEqual(expected, actual)

        params = {"inputs": {
            "f": PrimitiveParameterModel(
                name="f", type="string", value="filename"
            ),
        }}
        expected = "test -t tab filename"
        actual = builder.build_command(**params)
        self.assertEqual(expected, actual)

    def test_build_command_data(self):
        self._model.inputs = {
            "d": InputModel(name="d", type="Data", prefix='-d'),
            "e": InputModel(name="e", type="Data", prefix='-e', value={
                "name": "e",
                "url": "/path/to/.env",
                "path": ".env"
            }),
            "f": InputModel(name="f", type="Data"),
        }
        self._model.parameters["command"] = (
            "test $(inputs.d) $(inputs.e) $(inputs.f)"
        )
        dir_path = 'path/to/dirname'
        file_path = 'path/to/filename'
        env_path = 'path/to/file.env'
        builder = PythonRunner(self._model)

        params = {"inputs": {}}
        expected = f"test -e /s3/.env"
        actual = builder.build_command(**params)
        self.assertEqual(expected, actual)

        params = {"inputs": {
            "d": DataParameterModel(
                name="d", url=f'http://test.net', path=dir_path
            ),
            "e": DataParameterModel(
                name="e", url=f'http://test.net', path=env_path
            ),
            "f": DataParameterModel(
                name="f", url=f'http://test.net', path=file_path
            ),
        }}
        expected = f"test -d /s3/{dir_path} -e /s3/{env_path} /s3/{file_path}"
        actual = builder.build_command(**params)
        self.assertEqual(expected, actual)

    def test_build_dockerfile(self):
        expected = (
            "FROM --platform=linux/amd64 python:3.13-slim\n"
            "COPY requirements.txt /root/\n"
            "ENV MY_ENV=test\n"
            "RUN apt-get update ;\\\n"
            "\tapt-get install --no-install-recommends -y g++=4:12.2.0-3 "
            "gdal-bin libgdal-dev ;\\\n"
            "\tapt-get clean ;\\\n"
            "\trm -rf /var/apt/lists/* ;\\\n"
            "\tpip install --no-cache-dir -r /root/requirements.txt ;\\\n"
            "\tpip install --no-cache-dir drb-image-sentinel2==1.0.0 "
            "numpy awscli==1.32.20\n"
            "COPY . /delta\n"
            "WORKDIR /delta\n"
        )
        self._model.parameters["pythonVersion"] = "3.13"
        self._model.parameters["aptRequirements"] = [
            {"name": "g++", "version": "4:12.2.0-3"},
            {"name": "gdal-bin"},
            {"name": "libgdal-dev"},
        ]
        self._model.parameters["pipRequirementFiles"] = [
            "requirements.txt",
        ]
        self._model.parameters["pipRequirements"] = [
            {"name": "drb-image-sentinel2", "version": "1.0.0"},
            {"name": "numpy"},
        ]
        self._model.parameters["environment"] = {
            "MY_ENV": "test",
        }
        builder = PythonRunner(self._model)
        actual = builder.generate_dockerfile()
        self.assertEqual(expected, actual)



import unittest
import os

from delta.core import DeltaCore
from delta.vcs.handler import DeltaGitHandler


class TestCore(unittest.TestCase):
    def test_list(self):
        with DeltaCore() as core:
            try:
                core.list('url', 'token')
            except Exception:
                pass



from delta.store.interfaces import IFetcher
from delta.store.interfaces import IDataIO


class FetcherAllImpl(IFetcher):
    def __init__(self, dataIO: IDataIO):
        super().__init__(dataIO)

    def fetch(self):
        """
        Read and write data using DataIO.
        """
        data = self._dataIO._read()
        self._dataIO._write(data)



import unittest
import os

from delta.drive import DeltaDrive, DeltaTypeError, DeltaIllegalError
from delta.drive.interfaces import IDrive


class TestDrive(unittest.TestCase):

    def test_drive_tag(self):
        drive = DeltaDrive()
        with self.assertRaises(DeltaTypeError):
            drive.init(10)
        drive.init('test_drive_tag')
        drive.repo_directory = 'test_drive_tag'
        tags = drive.tag(list_tags=True, delete=True)
        self.assertEqual([], tags)
        with self.assertRaises(DeltaTypeError):
            drive.tag(tag_name=1)
        with self.assertRaises(DeltaTypeError):
            drive.tag(tag_name=None)
        with self.assertRaises(DeltaIllegalError):
            drive.tag(tag_name='')

        drive.tag(tag_name="toto")
        tags = drive.tag(list_tags=True)
        self.assertEqual(['toto'], tags)
        with self.assertRaises(DeltaTypeError):
            drive.tag(tag_name="toto", delete='error')
        drive.tag(tag_name="toto", delete=True)

    def test_drive_commit(self):
        drive = DeltaDrive()
        drive.init('test_drive_commit')
        drive.repo_directory = 'test_drive_commit'
        with self.assertRaises(DeltaTypeError):
            drive.commit(1)
        with self.assertRaises(DeltaIllegalError):
            drive.commit('')
        drive.commit('Hello world!')

    def test_drive_pull(self):
        drive = DeltaDrive()
        drive.init('test_drive_pull')
        drive.repo_directory = 'test_drive_pull'
        with self.assertRaises(DeltaTypeError):
            drive.pull(1)
        try:
            drive.pull()
        except Exception:
            pass

    def test_drive_push(self):
        drive = DeltaDrive()
        drive.init('test_drive_push')
        drive.repo_directory = 'test_drive_push'
        with self.assertRaises(DeltaTypeError):
            drive.push(remote=1, ref_name=1)
        with self.assertRaises(DeltaTypeError):
            drive.push(remote='toto', ref_name=1)
        try:
            drive.push(remote='toto', ref_name='test')
        except Exception:
            pass

    def test_drive_branch(self):
        drive = DeltaDrive()
        drive.init('test_drive_branch')
        drive.repo_directory = 'test_drive_branch'
        branches = drive.branch()
        self.assertEqual(['master'],  branches)
        with self.assertRaises(DeltaTypeError):
            drive.branch(branch_name=1, delete=True)
        with self.assertRaises(DeltaTypeError):
            drive.branch(branch_name="toto", delete=1)
        drive.branch(branch_name='toto')
        self.assertEqual(['master', 'toto'], drive.branch())
        drive.branch(branch_name='toto', delete=True)
        self.assertEqual(['master'], drive.branch())

    def test_drive_checkout(self):
        drive = DeltaDrive()
        drive.init('test_drive_checkout')
        drive.repo_directory = 'test_drive_checkout'
        self.assertEqual(['master'], drive.branch())
        with self.assertRaises(DeltaTypeError):
            drive.checkout(1)
        drive.checkout('toto')
        self.assertEqual(['master', 'toto'], drive.branch())

    def test_drive_reset(self):
        drive = DeltaDrive()
        drive.init('test_drive_reset')
        drive.repo_directory = 'test_drive_reset'
        drive.reset()

    def test_drive_add_dependency(self):
        drive = DeltaDrive()
        drive.init('test_drive_add_dependency')
        drive.repo_directory = 'test_drive_add_dependency'
        drive.add_dependency()

    def test_drive_add_resource(self):
        drive = DeltaDrive()
        drive.init('test_drive_add_resource')
        drive.repo_directory = 'test_drive_add_resource'
        with self.assertRaises(Exception):
            drive.add_resource(name=1, path='./goodurl/test', download=True)
        with self.assertRaises(Exception):
            drive.add_resource(name='resource_name', path=1, download=True)

    def test_drive_check(self):
        drive = DeltaDrive()
        drive.init('test_drive_check')
        drive.repo_directory = 'test_drive_check'
        self.assertTrue(drive.check())

    def test_drive_fetch(self):
        drive = DeltaDrive()
        drive.init('test_drive_fetch')
        drive.repo_directory = 'test_drive_fetch'
        with self.assertRaises(DeltaTypeError):
            drive.fetch(1)
        try:
            drive.fetch('origin')
        except Exception:
            pass

    def test_drive_interface(self):
        drive = IDrive()
        with self.assertRaises(NotImplementedError):
            drive.add_dependency()

        with self.assertRaises(NotImplementedError):
            drive.add_resource()

        with self.assertRaises(NotImplementedError):
            drive.branch()

        with self.assertRaises(NotImplementedError):
            drive.branch(branch_name='toto')

        with self.assertRaises(NotImplementedError):
            drive.branch(branch_name='toto', delete=True)

        with self.assertRaises(NotImplementedError):
            drive.check()

        with self.assertRaises(NotImplementedError):
            drive.checkout('branch')

        with self.assertRaises(NotImplementedError):
            drive.clone('url', 'path')

        with self.assertRaises(NotImplementedError):
            drive.commit('message')

        with self.assertRaises(NotImplementedError):
            drive.fetch('origin')

        with self.assertRaises(NotImplementedError):
            drive.init('path')

        with self.assertRaises(NotImplementedError):
            drive.pull('origin')

        with self.assertRaises(NotImplementedError):
            drive.push()

        with self.assertRaises(NotImplementedError):
            drive.reset()

        with self.assertRaises(NotImplementedError):
            drive.status()

        with self.assertRaises(NotImplementedError):
            drive.sync()

        with self.assertRaises(NotImplementedError):
            drive.tag()

    def test_execute_git_command(self):
        _, status = DeltaDrive.execute_git_command(
            "init",
            "test_execute_git_command")
        self.assertEqual(0, status)
        _, status = DeltaDrive.execute_git_command(
            "checkout",
            "-b",
            "toto",
            path="test_execute_git_command")
        self.assertEqual(0, status)

        _, status = DeltaDrive.execute_git_command(
            "not",
            "a",
            "git",
            "command")
        self.assertEqual(1, status)

        _, status = DeltaDrive.execute_git_command(
            "not",
            "a",
            "git",
            "command",
            path="test_execute_git_command")
        self.assertEqual(1, status)

        with self.assertRaises(TypeError):
            DeltaDrive.execute_git_command(
                "not",
                "a",
                "git",
                "command",
                path=1)



from abc import ABC


class IData(ABC):
    # Contient source et dest
    def __init__(self):
        self._info: dict = {}

    @property
    def info(self) -> dict:
        return self._info

    def add_info(self, **kwargs):
        for key, value in kwargs.items():
            self._info[key] = value

    def remove_info(self, key):
        return self._info.pop(key, None)



import logging
import os
import tarfile
import tempfile
from typing import Callable

import docker
from git import Git

from delta.drive.interfaces import IDrive
from delta.manifest import manifest
from delta.manifest.parser import Model, parse
from delta.run.job.builder import JobBuilder
from delta.run.job.builder.python_builder import PythonRunner
from delta.store.impl import StoreImpl
from delta.vcs.handler import DeltaGitHandler


class DeltaTypeError(Exception):
    pass


class DeltaIllegalError(Exception):
    pass


class DeltaDrive(IDrive):

    def __init__(self):
        self._repo_directory = "."

    @property
    def repo_directory(self):
        return self._repo_directory

    @repo_directory.setter
    def repo_directory(self, directory):
        self._repo_directory = directory

    def clone(self, url: str, path_to_save: str = repo_directory):
        r"""
        Clone a repository.

        :param url:
        URL where the repository is located.
        :type url: ``str``

        :param path_to_save:
        The path where we want to save locally the repository.
        :type path_to_save: ``str``
        """
        if not isinstance(path_to_save, str):
            raise DeltaTypeError("Wrong type for path_to_save. Need a str.")
        with DeltaGitHandler() as repo:
            repo.clone(url, path_to_save)

    def init(self, path_to_save: str):
        r"""
        Initialize a new repository.

        :param path_to_save:
        Path where we want to save locally the repository.
        :type path_to_save: ``str``
        """
        if not isinstance(path_to_save, str):
            raise DeltaTypeError("Wrong type for path_to_save. Need a str.")
        with DeltaGitHandler() as repo:
            repo.create(path_to_save)

    def status(self):
        pass

    def tag(self, **kwargs):
        r"""
        Can remove, list and create tags.

        :Keyword Arguments:
        * *list_tags* (``bool``) --
            Flag to list all tags in repo.
        * *tag_name* (``str``) --
            Tag name needs to be created or removed.
        * *delete* (``bool``) --
            Flag to delete or not the tag.
        * *message* (``str``) --
            Message when we create the new tag.
        """
        list_tags = kwargs.get("list_tags", False)
        if list_tags:
            with DeltaGitHandler() as repo:
                repo.load_local_repo(self.repo_directory)
                return repo.list_tag()
        else:
            tag_name = kwargs.get("tag_name")
            if not isinstance(tag_name, str):
                raise DeltaTypeError("Wrong type for tag_name. Need a str.")
            if tag_name == "":
                raise DeltaIllegalError("Illegal tag name")
            delete = kwargs.get("delete", False)
            if not isinstance(delete, bool):
                raise DeltaTypeError("Wrong type for delete. Need a bool")
            with DeltaGitHandler() as repo:
                repo.load_local_repo(self.repo_directory)
                if delete:
                    repo.remove_tag(tag_name)
                else:
                    message = kwargs.get("message", "")
                    if not isinstance(message, str):
                        raise DeltaTypeError(
                            "Wrong type for message." "Need a str."
                        )
                    repo.tag(tag_name, message)

    def commit(self, message: str):
        r"""
        Save all modification.

        :param message:
        Message associated of the commit.
        :type message: ``str``
        """
        if not isinstance(message, str):
            raise DeltaTypeError("Wrong type for message. Need a str.")
        if message == "":
            raise DeltaIllegalError(
                "Wrong message for your commit. " "Need a non-empty message."
            )
        with DeltaGitHandler() as repo:
            repo.load_local_repo(self.repo_directory)
            repo.commit(message)

    def pull(self, remote_name: str):
        r"""
        Fetch distant references and all modification will
        be integrated by the current local branch.

        :param remote_name:
        Name of the remote.
        :type remote_name: ``str``
        """
        if remote_name is not None:
            if not isinstance(remote_name, str):
                raise DeltaTypeError("Wrong type for remote_name. Need a str.")
        with DeltaGitHandler() as repo:
            repo.load_local_repo(self.repo_directory)
            repo.pull(remote_name)

    def push(self, **kwargs):
        r"""
        Push to origin.
        If no remote has been specified, 'origin' will
        be used by default.

        :param \**kwargs:
        See below

        :Keyword Arguments:
        * *remote* (``str``) --
            Remote name.
        * *ref_name* (``str``) --
            Tag name needs to be pushed.
        """
        remote_name = kwargs.get("remote", None)
        if remote_name is not None:
            if not isinstance(remote_name, str):
                raise DeltaTypeError("Wrong type for origin_name. Need a str.")
        ref_name = kwargs.get("ref_name", None)
        if ref_name is not None:
            if not isinstance(ref_name, str):
                raise DeltaTypeError(
                    "Wrong type for reference name. " "Need a str"
                )
        with DeltaGitHandler() as repo:
            repo.load_local_repo(self.repo_directory)
            repo.push(remote_name, ref_name)

    def branch(self, **kwargs):
        r"""
        List all branches available in a repo.
        Also, create and delete a branch.

        :param \**kwargs:
        See below

        :returns:
        No arguments or there isn't `branch_name`,
        will return a list of all branches available.
        Or return `None`.

        :Keyword Arguments:
        * *branch_name* (``str``) --
            Branch name you want to create or delete.
        * *delete* (``bool``) --
            Flag to delete branch.
        """
        branch_name = kwargs.get("branch_name", None)
        if branch_name is None:
            with DeltaGitHandler() as repo:
                repo.load_local_repo(self.repo_directory)
                return repo.list_branches()
        delete = kwargs.get("delete", False)
        if not isinstance(branch_name, str):
            raise DeltaTypeError("Wrong type for branch_name. Need a str.")
        if not isinstance(delete, bool):
            raise DeltaTypeError("Wrong type for delete. Need a bool.")
        with DeltaGitHandler() as repo:
            repo.load_local_repo(self.repo_directory)
            if delete:
                repo.remove_branch(branch_name)
            else:
                repo.branch(branch_name)

    def checkout(self, ref_name: str):
        r"""
        Switch to a reference of your repo (tag, commit, or branch).

        :param ref_name:
        Name of the reference.
        :type ref_name: ``str``
        """
        if not isinstance(ref_name, str):
            raise DeltaTypeError("Wrong type for ref_name. Need a str.")
        with DeltaGitHandler() as repo:
            repo.load_local_repo(self.repo_directory)
            repo.checkout(ref_name)

    def reset(self):
        pass

    def add_dependency(self):
        pass

    def add_resource(self, **kwargs):
        r"""
        Add a resource to manifest and resources folder.
        So it will fetch data and write file to resources folder.

        :Keyword Arguments:
        * *name* (``str``) --
            Resource name.
        * *path* (``str``) --
            Data where we can find the resource file
            (it could be an URL or a filepath).
        * *download* (``boolean``) --
            A boolean that tell if the data will be downloaded
            in the resources folder.
        """
        store = StoreImpl()
        store.add_resource(**kwargs)

    def delete_resource(self, **kwargs):
        store = StoreImpl()
        store.remove_resource(**kwargs)

    def get_resources(self):
        store = StoreImpl()
        return store.get_resources()

    def check(self) -> bool:
        path = os.path.join(
            os.path.abspath(self.repo_directory), "manifest.json"
        )
        data = manifest.read_manifest(path)
        return manifest.check_manifest(data)

    def fetch(self, remote_name: str):
        r"""
        Update all references of a remote.
        Equivalent of git fetch.

        :param remote_name:
        Name of the remote.
        :type remote_name: ``str``
        """
        if remote_name is not None:
            if not isinstance(remote_name, str):
                raise DeltaTypeError("Wrong type for remote_name. Need a str.")
        with DeltaGitHandler() as repo:
            repo.load_local_repo(self.repo_directory)
            repo.fetch(remote_name)

    def sync(self):
        manifest_path = os.path.join(
            os.path.abspath(self.repo_directory), "manifest.json"
        )
        _manifest = manifest.read_manifest(manifest_path)
        if "resources" in _manifest:
            for item in _manifest["resources"]["items"]:
                self.add_resource(
                    name=item["name"], path=item["source_url"], download=True
                )

    @staticmethod
    def execute_git_command(*args, **kwargs):
        """
        Execute a git command.

        Available kwargs :
        * *path* (``str`` or ``None``) --
            Corresponds to the repository path.

        :return (str(stdout | stderr), (int(status)))

        Examples :
        * execute_git_command("commit", "-m", "this is a commit")
        * execute_git_command("log", path="repository_path")
        * execute_git_command("status")
        """
        path = kwargs.get("path", None)
        if not isinstance(path, (str, type(None))):
            raise TypeError("Error : path must be None or a str")
        status, out, err = Git(path).execute(
            ["git", *args],
            with_exceptions=False,
            with_extended_output=True,
            stdout_as_string=True,
        )
        res = err if (status != 0) else out
        return res, status

    @staticmethod
    def get_runner_factory(kind: str) -> Callable[[Model], JobBuilder]:
        if kind == "python":
            return PythonRunner
        raise ValueError(f"Unknown runner kind '{kind}'")

    @staticmethod
    def generate_docker_build_context(
        path: str, dockerfile_content: str
    ) -> str:
        with tempfile.NamedTemporaryFile(suffix=".tar", delete=False) as ctx:
            with tempfile.NamedTemporaryFile() as dockerfile:
                dockerfile.write(dockerfile_content.encode(encoding="utf-8"))
                dockerfile.flush()
                with tarfile.open(ctx.name, "w") as tar:
                    tar.add(path, arcname=".")
                    tar.add(
                        dockerfile.name, arcname="Dockerfile", recursive=False
                    )
            return ctx.name

    def docker_build(
        self,
        version: str = "dev",
        registry: str = None,
        no_cache: bool = False,
    ):
        manifest_json = parse(
            os.path.join(self._repo_directory, "manifest.json")
        )
        if not manifest_json.models:
            logging.info(f"No models images to build")
            return
        for name, model in manifest_json.models.items():
            logging.info(f"Building image for '{name}' model.")
            # retrieve runner
            runner = self.get_runner_factory(model.type)(model)
            # generate dockerfile via the runner
            dockerfile_str = runner.generate_dockerfile()
            # generate docker build context
            context = self.generate_docker_build_context(
                os.path.join(self._repo_directory, model.path), dockerfile_str
            )
            # perform docker build
            try:
                with open(context, "rb") as ctx:
                    tag = f"{manifest_json.name}/{name}:{version}"
                    if registry is not None:
                        tag = f"{registry}/{tag}"
                    cli = docker.from_env()
                    cli.images.build(
                        custom_context=True,
                        rm=True,
                        fileobj=ctx,
                        tag=tag,
                        platform="linux/amd64",
                        nocache=no_cache,
                    )
                    logging.info(f"'{name}' model image successfully built.")
            except docker.errors.DockerException as ex:
                logging.error("Error during docker build")
                raise ex
            finally:
                # remove docker build
                os.remove(context)



from delta.runners.run import DeltaRun, get
from delta.exceptions.runners import DeltaRunnerError, DeltaRunnerNotFound


def create(model: dict) -> DeltaRun:
    if 'type' not in model:
        raise DeltaRunnerError('Missing required runner type: "type"')

    runner_name = model['type']
    try:
        runner_factory_class = get(runner_name)
    except KeyError:
        raise DeltaRunnerNotFound(f'Unknown runner type: "{runner_name}"')

    runner = runner_factory_class(model.get('parameters'))
    return runner



import re
from typing import Annotated

from pydantic import (Field, HttpUrl, PostgresDsn,
                      UrlConstraints, field_validator)
from pydantic_core import MultiHostUrl
from pydantic_settings import BaseSettings, SettingsConfigDict
from rsa import PublicKey, PrivateKey


SQLiteUrl = Annotated[
    MultiHostUrl,
    UrlConstraints(
        host_required=False,
        allowed_schemes=['sqlite', 'sqlite3'],
    )
]


def format_pem_key(key: str, key_type: str) -> str:

    pem_header_footer_regex = (rf'(-----BEGIN RSA '
                               rf'{key_type} KEY-----)(.+)(-----END RSA '
                               rf'{key_type} KEY-----)')
    matches = re.match(pem_header_footer_regex, key)

    if matches:
        pem_content = matches.group(2)
        formatted_pem_content = '\n'.join(
            [pem_content[i:i+64] for i in range(0, len(pem_content), 64)])
        return (f"{matches.group(1)}\n"
                f"{formatted_pem_content}\n{matches.group(3)}")
    else:
        raise ValueError(f"The {key_type.lower()} key string does not "
                         f"match the expected PEM format.")


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        case_sensitive=False,
        env_prefix='DELTA_',
    )

    # Deltatwin Run
    gss_api_url: HttpUrl
    run_limit: int = Field(default=10)
    eviction_active: bool = Field(default=False)
    eviction_keep_period: int = Field(default=48)
    database_url: SQLiteUrl | PostgresDsn = Field(
        default="sqlite:///.delta-run.db?check_same_thread=false"
    )
    database_show_sql: bool = Field(default=False)
    page_limit: int = Field(default=100)
    public_key: str
    private_key: str

    # DeltaTwin container image registry
    image_repo_hostname: str
    image_repo_username: str
    image_repo_password: str

    # Keycloak
    socketio_adapter_url: str
    keycloak_jwks_url: HttpUrl

    # DeltaTwin Run output storage
    s3_endpoint: str
    s3_region: str
    s3_access_key: str
    s3_secret_access_key: str
    s3_bucket: str

    # kubernetes executor
    k8s_context: str
    k8s_namespace: str
    k8s_cluster_name: str
    k8s_cluster_cert_auth: str
    k8s_cluster_server: str
    k8s_user_name: str
    k8s_user_cli_cert: str
    k8s_user_cli_key: str

    @field_validator('public_key')
    def decode_public_key(cls, v: str) -> PublicKey:
        formatted_pem_key = format_pem_key(v, 'PUBLIC')
        return PublicKey.load_pkcs1(
            formatted_pem_key.encode('utf-8'), format='PEM')

    @field_validator('private_key')
    def decode_private_key(cls, v: str) -> PrivateKey:
        formatted_pem_key = format_pem_key(v, 'PRIVATE')
        return PrivateKey.load_pkcs1(
            formatted_pem_key.encode('utf-8'), format='PEM')



from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from delta.run.config import Settings


engine = create_engine(
    url=str(Settings().database_url),
    echo=bool(Settings().database_show_sql),
)

Session = sessionmaker(bind=engine, autoflush=True)



import unittest
import delta.runners.factory as runner_factory
from delta.exceptions.runners import DeltaRunnerNotFound, DeltaRunnerError
from delta.runners.tests import AbstractDeltaRunnerTest


class TestRunnerFactory(unittest.TestCase):
    def test_create(self):
        model = {"type": 'runner_a'}
        runner = runner_factory.create(model)
        self.assertIsNotNone(runner)
        self.assertIsInstance(runner, AbstractDeltaRunnerTest)
        self.assertEqual('A', runner.value)

        model = {"type": 'runner_b'}
        runner = runner_factory.create(model)
        self.assertIsNotNone(runner)
        self.assertIsInstance(runner, AbstractDeltaRunnerTest)
        self.assertEqual('B', runner.value)

        model = {"type": "foobar"}
        with self.assertRaises(DeltaRunnerNotFound):
            runner = runner_factory.create(model)

        model = {"parameters": {"foo": "bar"}}
        with self.assertRaises(DeltaRunnerError):
            runner = runner_factory.create(model)



from abc import ABC, abstractmethod


class IStore(ABC):
    @abstractmethod
    def add_resource(self, **kwargs):
        raise NotImplementedError()

    @abstractmethod
    def remove_resource(self, **kwargs):
        raise NotImplementedError()

    @abstractmethod
    def get_resources(self):
        raise NotImplementedError()



import unittest
import os

from delta.vcs.handler.delta_git_handler import DeltaGitHandler, DeltaException


class TestLoadLocalRepo(unittest.TestCase):
    def test_load_not_local_repo(self):
        gh = DeltaGitHandler()
        with self.assertRaises(DeltaException):
            gh.load_local_repo('test_load_not_local_repo_')
        self.assertTrue(gh._repo is None)
        self.assertTrue(gh._path is None)
        self.assertTrue(gh._initialize_flag is False)

    def test_load_local_repo(self):
        gh = DeltaGitHandler()
        self.assertIsNone(gh.create('test_load_local_repo1'))
        self.assertIsNone(gh.create('test_load_local_repo2'))
        self.assertIsNotNone(gh._repo)
        self.assertEqual(
            gh._path,
            os.path.join(
                os.path.abspath('.'),
                'test_load_local_repo2'
                )
            )
        self.assertTrue(gh._initialize_flag is True)

        self.assertIsNone(gh.load_local_repo('test_load_local_repo1'))

        self.assertIsNotNone(gh._repo)
        self.assertEqual(gh._path,
                         os.path.join(os.path.abspath('.'),
                                      'test_load_local_repo1'))
        self.assertTrue(gh._initialize_flag is True)



import os
from unittest import TestCase

from delta.manifest import parse
from delta.manifest.parser import Copyright, License, Manifest, Model, Resource


def _json_sample_path():
    return os.path.join(os.path.dirname(__file__))


class TestParse(TestCase):
    path_full = None
    path_minimum = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.path_full = os.path.join(
            _json_sample_path(), 'test_manifest_full.json')
        cls.path_minimum = os.path.join(
            _json_sample_path(), 'test_manifest_minimum.json')

    def test_description(self):
        manifest = parse(self.path_full)
        self.assertEqual(manifest.name, 'Delta Twin name')
        self.assertEqual(manifest.description, 'Delta Twin Description')
        expected_license = License(
            name='My own licence',
            description='description of the licences',
            url='http://path_to_the_licence',
            copyrights=[
                Copyright(company='Gael Systems',
                          years=[2021, 2022, 2023, 2024]),
                Copyright(company='ESA/ESRIN', years=[2021])
            ]
        )
        self.assertEqual(expected_license, manifest.license)

    def test_resources(self):
        # check on minimal manifest
        resources = parse(self.path_minimum).resources
        self.assertIsInstance(resources, dict)
        self.assertEqual({}, resources)

        # check on complete manifest
        resources = parse(self.path_full).resources
        self.assertIsInstance(resources, dict)
        self.assertEqual(len(resources), 3)

        name = ("S2B_MSIL1C_20210211T095029_N0209_R079_T33TWF"
                "_20210211T111431.zip")
        resource = resources.get(name)
        self.assertIsNotNone(resource)
        self.assertIsInstance(resource, Resource)
        self.assertEqual(resource.name, name)
        resource_value = ("${remote_service}/Product"
                          "('93d1ea14-e76b-4268-8332-8ab5070aa0cc')")
        self.assertEqual(resource_value, resource.value)

    def test_models(self):
        models = parse(self.path_minimum).models
        self.assertIsInstance(models, dict)
        self.assertEqual({}, models)

        models = parse(self.path_full).models
        self.assertIsInstance(models, dict)
        self.assertNotEqual({}, models)
        expected_model = Model(
            path='models',
            type='nifi',
            parameters={
                'parameters1': 10,
                'parameters2': 'hello',
                'parameters3': {
                    'Hello': 'World',
                    'Dict': {
                        'subparameters1': 10,
                        'subparameters2': 'test',
                    }
                },
                'parameters4': ['Hello', 'World'],
            }
        )
        self.assertEqual(expected_model, models.get("single_model"))

    def test_manifest(self):
        manifest = parse(self.path_full)
        self.assertIsInstance(manifest, Manifest)

        self.assertIsInstance(manifest.name, str)
        self.assertEqual('Delta Twin name', manifest.name)

        self.assertIsInstance(manifest.description, str)
        self.assertEqual('Delta Twin Description', manifest.description)

        self.assertIsInstance(manifest.license, License)

        self.assertIsInstance(manifest.resources, dict)
        self.assertEqual(3, len(manifest.resources))

        self.assertIsInstance(manifest.models, dict)
        self.assertEqual(1, len(manifest.models))
        self.assertIsNotNone(manifest.models.get("single_model"))

        self.assertIsInstance(manifest.dependencies, dict)
        self.assertEqual(0, len(manifest.dependencies))



from delta.runners.run import DeltaRun


class AbstractDeltaRunnerTest(DeltaRun):
    def __init__(self, parameters: dict, value: str):
        super().__init__(parameters)
        self.__value = value

    def configure(self, **kwargs):
        pass

    def start(self, **kwargs):
        pass

    def stop(self, **kwargs):
        pass

    def resume(self, **kwargs):
        pass

    def monitor(self, **kwargs):
        pass

    @property
    def value(self) -> str:
        return self.__value

    @property
    def get_parameters(self) -> dict:
        return self._parameters


class RunnerA(AbstractDeltaRunnerTest):
    def __init__(self, parameters: dict):
        super().__init__(parameters, 'A')


class RunnerB(AbstractDeltaRunnerTest):
    def __init__(self, parameters: dict):
        super().__init__(parameters, 'B')



import dataclasses
import io
import re
from typing import List

from delta.manifest.parser import Model
from delta.run.job.builder import JobBuilder


PYTHON_IMAGE_HASH_MAPPING = {
    # bookworm
    "3.10": (
        "sha256:"
        "a4617b0df2fea9da7ff505bf68b023c467c8d4b21e59b4b8e3de40dab4b0bf7f"
    ),
    "3.11": (
        "sha256:"
        "50068b28f3105e4a3116e091993304eccb38895d676c8dae174c5b01a4d30dfb"
    ),
    # trixie
    "3.12": (
        "sha256:"
        "d86b4c74b936c438cd4cc3a9f7256b9a7c27ad68c7caf8c205e18d9845af0164"
    ),
    "3.13": (
        "sha256:"
        "85dfbf1b566b7addfe645faea9938e81a0a01a83580b0ea05fb23706357d77fb"
    ),
    "3.14": (
        "sha256:"
        "9813eecff3a08a6ac88aea5b43663c82a931fd9557f6aceaa847f0d8ce738978"
    ),
}


@dataclasses.dataclass
class SoftwarePackage:
    name: str
    version: str | None = None

    def render(self, sep: str = "=") -> str:
        return f"{self.name}{sep}{self.version}" if self.version else self.name


class PythonRunner(JobBuilder):
    def __init__(self, model: Model):
        self.model = model

    @staticmethod
    def __build_run_command(
            apt_requirements: List[SoftwarePackage],
            pip_requirements: List[SoftwarePackage],
            pip_requirement_files: List[str],
    ) -> str:
        run = []
        if apt_requirements:
            segment = " ".join([x.render("=") for x in apt_requirements])
            run.append("apt-get update")
            run.append(f"apt-get install --no-install-recommends -y {segment}")
            run.append("apt-get clean")
            run.append("rm -rf /var/apt/lists/*")

        if pip_requirement_files:
            segment = [f"-r /root/{f}" for f in pip_requirement_files]
            run.append(f"pip install --no-cache-dir {' '.join(segment)}")

        if pip_requirements:
            segment = " ".join([x.render("==") for x in pip_requirements])
            run.append(f"pip install --no-cache-dir {segment}")
        return " &&\\\n\t".join(run)

    def generate_dockerfile(self) -> str:
        params = self.model.parameters
        python_version = params['pythonVersion']
        if not PYTHON_IMAGE_HASH_MAPPING.get(python_version):
            raise ValueError(
                f"Python version {python_version} is not supported"
            )
        with io.StringIO() as buf:
            buf.write(
                "FROM --platform=linux/amd64 "
                f"python@{PYTHON_IMAGE_HASH_MAPPING[python_version]}\n"
            )
            # copying requirement files if necessary
            if params.get("pipRequirementFiles"):
                requirement_files = " ".join(params["pipRequirementFiles"])
                buf.write(f"COPY {requirement_files} /root/\n")
            # prepare command
            apt_requirements = [
                SoftwarePackage(**e) for e in params.get("aptRequirements", [])
            ]
            pip_requirements = [
                SoftwarePackage(**e) for e in params.get("pipRequirements", [])
            ]
            # FIXME remove this when no additional copy will no be necessary
            pip_requirements.append(SoftwarePackage(
                name="awscli", version="1.32.20"
            ))
            if params.get("environment"):
                for key, value in params.get("environment", {}).items():
                    buf.write(f'ENV {key}={value}\n')
            run_command = self.__build_run_command(
                apt_requirements=apt_requirements,
                pip_requirements=pip_requirements,
                pip_requirement_files=params.get("pipRequirementFiles", []),
            )
            if run_command:
                buf.write(f"RUN {run_command}\n")
            buf.write("COPY . /delta\n")
            buf.write("WORKDIR /delta\n")
            return buf.getvalue()

    def build_command(self, secure: bool = False, **kwargs) -> str:
        cmd_pattern = self.model.parameters["command"]
        evaluated_cmd = []
        for cmd_segment in cmd_pattern.split(" "):
            m = re.match(
                r"\$\(([a-z][A-Za-z0-9_-]*)(\.[a-z][A-Za-z0-9_-]*)*\)",
                cmd_segment
            )
            if m:
                keys = m.group()[2:-1].split(".")
                model_part = getattr(self.model, keys[0])
                model_part = model_part[keys[1]]
                value = kwargs[keys[0]].get(keys[1])
                placeholder = self.placeholder_factory(model_part, value)
                evaluated_cmd.append(placeholder.evaluate(secure))
            else:
                evaluated_cmd.append(cmd_segment)
        return " ".join(evaluated_cmd).strip().replace("  ", " ")



import inspect
import logging
import platform
import importlib.metadata
from abc import ABC, abstractmethod
from typing import Callable, Iterable, Optional
from delta.exceptions.runners import DeltaRunnerNotFound


_entry_point_name = 'delta.runner'
_runners = None


# Define the DeltaRun interface
class DeltaRun(ABC):

    def __init__(self, parameters: Optional[dict] = None):
        self._parameters = parameters or {}

    @abstractmethod
    def configure(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def start(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def stop(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def resume(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def monitor(self, **kwargs):
        raise NotImplementedError


def _get_entry_points(name: str) -> Iterable[importlib.metadata.EntryPoint]:
    version = platform.python_version()
    major, minor, patch = map(lambda x: int(x), version.split('.'))
    eps = importlib.metadata.entry_points()

    # Python 3.8 & 3.9 importlib.metadata.entry_points() -> dict
    if minor == 8 or minor == 9:
        if name in eps:
            return eps[name]

    # Python 3.10+ importlib.metadata.entry_points() -> Collection
    elif minor >= 10:
        if name in eps.groups:
            return eps.select(group=name)

    return []


def _filter_runner_by_name(name: str):
    """
    Returns specific lambda checking if the given object is a class, and it
    names as the given name.
    """
    return lambda obj: inspect.isclass(obj) and obj.__name__ == name


def _load_runners():
    """
    Loads runners defined in the Python environment via the entry point
    mechanism.
    """
    for entry_point in _get_entry_points(_entry_point_name):
        if entry_point.name in _runners:
            logging.warning(
                f'Runner({entry_point.name},{entry_point.value}) skipped '
                f'caused by runner name ({entry_point.name}) already used')
            continue

        module_path, class_name = entry_point.value.split(':')
        try:
            module = importlib.import_module(module_path)

            members = inspect.getmembers(module,
                                         _filter_runner_by_name(class_name))
            if len(members) != 1:
                raise DeltaRunnerNotFound(
                    f'DeltaRun not found at: {entry_point.value}')

            _, obj = members[0]
            if not issubclass(obj, DeltaRun):
                raise TypeError(f'{entry_point.value} is not a DeltaRun')

            _runners[entry_point.name] = obj
        except ModuleNotFoundError:
            logging.error(f'module {module_path} not found')
        except Exception as ex:
            logging.error('an error occurred during loading runner '
                          f'"{entry_point.name}": {ex}')


def get(name: str) -> Callable:
    if name not in _runners:
        raise KeyError(f'Runner not found: {name}')
    return _runners[name]


if _runners is None:
    _runners = {}
    _load_runners()



import os.path
import unittest
from delta.manifest import parse
from delta.manifest.parser import (
    Copyright, Input, InputModel, License, Manifest, Model, Output,
    OutputModel, Resource
)


class TestParserV1v1(unittest.TestCase):
    full = None

    @classmethod
    def setUpClass(cls):
        cls.full = os.path.join(os.path.dirname(__file__),
                                'test_manifest_full.json')

    def test_parse(self):
        expected_name = "testDeltaTwin"
        expected_description = 'Any description'

        expected_license = License(
            name="LGPLv3",
            description="license description",
            url="https://www.gnu.org/licenses/gpl-3.0.txt",
            copyrights=[Copyright(company="Gael Systems", years=[2023])]
        )

        resources = [
            Resource(
                name="aData",
                type="Data",
                description="resource data description",
                value="https://domain.net/path/data"
            ),
            Resource(
                name="aValue",
                type="string",
                value="Value"
            )
        ]
        expected_resource = {e.name: e for e in resources}

        inputs = [
            Input(name="in_data", type="Data"),
            Input(name="in_bool", type="boolean", value=False)
        ]
        expected_inputs = {e.name: e for e in inputs}

        outputs = [Output(name="out_data", type="Data")]
        expected_outputs = {e.name: e for e in outputs}

        json_formatter = Model(
            path="models/jsonFormatter",
            type="python",
            parameters={
                "image": "python:3.10",
                "cmd": "python -m json.tool"
            },
            inputs={
                "sortedKeys": InputModel(
                    name="sortedKeys",
                    type="boolean",
                    value=False,
                    prefix="--sorted-keys"
                ),
                "json": InputModel(
                    name="json",
                    type="Data"
                )
            },
            outputs={"out": OutputModel(name="out", type="stdout")}
        )
        band_selector = Model(
            path="models/band-selector",
            type="python",
            parameters={
                "image": "python:3.8",
                "requirementFiles": ["requirements.txt"],
                "pyFile": "band_extractor.py",
                "cmd": "python $(parameters.pyFile)"
            },
            inputs={
                "product": InputModel(
                    name="product",
                    type="Data"
                ),
                "bandName": InputModel(
                    name="bandName",
                    type="string"
                )
            },
            outputs={
                "selectedBand": OutputModel(
                    name="selectedBand",
                    type="Data",
                    glob="band/*"
                )
            }
        )
        expected_models = {
            "json-formatter": json_formatter,
            "band-selector": band_selector
        }

        expected_dependencies = [
            "test-deltatwin1==1.0.0",
            "test-deltatwin2==1.0.1"
        ]

        expected_manifest = Manifest(
            name=expected_name,
            description=expected_description,
            license=expected_license,
            resources=expected_resource,
            inputs=expected_inputs,
            outputs=expected_outputs,
            models=expected_models,
            dependencies=expected_dependencies
        )

        manifest = parse(self.full)
        self.assertIsNotNone(manifest)
        self.assertIsInstance(manifest, Manifest)
        self.assertEqual(expected_name, manifest.name)
        self.assertEqual(expected_description, manifest.description)
        self.assertEqual(expected_license, manifest.license)
        self.assertEqual(expected_resource, manifest.resources)
        self.assertEqual(expected_inputs, manifest.inputs)
        self.assertEqual(expected_outputs, manifest.outputs)
        self.assertEqual(expected_models, manifest.models)
        self.assertEqual(expected_dependencies, manifest.dependencies)
        self.assertEqual(expected_manifest, manifest)



from typing import Sequence

from sqlalchemy.orm import Session
from sqlalchemy import and_, func, select, update

from delta.run.api.model import RunContextUpdateModel, model_param_to_orm_param
from delta.run.db.orm import Parameter, ParameterKind, RunContext, RunStatus
from datetime import datetime, timedelta


def get_runs(
    session: Session,
    dt_id: str = None,
    owner: str = None,
    status: RunStatus = None,
    limit: int = 0,
    offset: int = 0,
) -> Sequence[RunContext]:
    stmt = session.query(RunContext)
    if dt_id is not None:
        stmt = stmt.filter(RunContext.deltatwin_id == dt_id)
    if owner is not None:
        stmt = stmt.filter(RunContext.owner == owner)
    if status is not None:
        stmt = stmt.filter(RunContext.status == status)
    if limit > 0:
        stmt = stmt.limit(limit)
    if offset > 0:
        stmt = stmt.offset(offset)
    return session.scalars(stmt).all()


def add_run(session: Session, run: RunContext) -> RunContext:
    session.add(run)
    return session.get_one(RunContext, run.id)


def update_run(session: Session, run_id, run: RunContextUpdateModel):
    db_run_ctx = get_run_by_id(session, run_id)
    if run.status:
        db_run_ctx.status = run.status
    if run.return_code:
        db_run_ctx.return_code = run.return_code
    if run.message:
        db_run_ctx.message = run.message
    if run.inputs:
        inputs = [
            model_param_to_orm_param(p, ParameterKind.INPUT)
            for p in run.inputs
        ]
        db_run_ctx.inputs = inputs
    if run.outputs:
        outputs = [
            model_param_to_orm_param(p, ParameterKind.OUTPUT)
            for p in run.outputs
        ]
        db_run_ctx.outputs = outputs


def get_run_by_id(session: Session, id_: str) -> RunContext:
    return session.get_one(RunContext, id_)


def get_runs_by_date(session: Session,
                     threshold_date: datetime) -> Sequence[RunContext]:
    stmt = session.query(RunContext).filter(
        RunContext.date_created < threshold_date)
    return session.scalars(stmt).all()


def delete_run(session: Session, run_id: str):
    run = get_run_by_id(session, run_id)
    if run:
        session.delete(run)
        session.commit()


def get_data_output(
    session: Session, run_id: str, name_output: str, type_output: str
) -> Sequence[Parameter]:
    stmt = session.query(Parameter)
    if run_id is not None:
        stmt = stmt.filter(Parameter.run_id == run_id)
    if name_output is not None:
        stmt = stmt.filter(Parameter.name == name_output)
    if type_output is not None:
        stmt = stmt.filter(Parameter.type == type_output)

    stmt = stmt.filter(Parameter.kind == ParameterKind.OUTPUT)

    return session.scalars(stmt).all()


def get_number_of_run_by_user(session: Session, user: str) -> int:
    now = datetime.now()
    threshold_date = now - timedelta(hours=24)
    stmt = (
        select(func.count())
        .select_from(RunContext)
        .where(
            and_(
                RunContext.date_created > threshold_date,
                RunContext.owner == user,
                RunContext.parent_run == "",
                RunContext.return_code != 410,
            )
        )
    )
    return session.scalar(stmt)


def set_running_runs_to_error_at_startup(session: Session):
    stmt = (
        update(RunContext)
        .where(RunContext.status == RunStatus.RUNNING)
        .values(
            status=RunStatus.ERROR,
            return_code=410,
            message="Unexpected error, please retry later.",
        )
    )
    session.execute(stmt)
    session.commit()



"""Update run status enum

Revision ID: 1ec1eabe45bf
Revises: 2e57fc6dd873
Create Date: 2024-04-16 10:55:22.821262

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '1ec1eabe45bf'
down_revision: Union[str, None] = '2e57fc6dd873'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    dialect = op.get_context().dialect.name
    match dialect:
        case 'postgresql':
            op.execute("ALTER TYPE runstatus RENAME VALUE 'SUCCEEDED' TO 'SUCCESS'")
        case _:
            op.execute("UPDATE runs SET status = 'SUCCESS' WHERE status = 'SUCCEEDED'")


def downgrade() -> None:
    dialect = op.get_context().dialect.name
    match dialect:
        case 'postgresql':
            op.execute("ALTER TYPE runstatus RENAME VALUE 'SUCCESS' TO 'SUCCEEDED'")
        case _:
            op.execute("UPDATE runs SET status = 'SUCCEEDED' WHERE status = 'SUCCESS'")



import asyncio
import collections
import datetime
import io
import logging
import os
import timeit
import uuid
from dataclasses import dataclass, field
from functools import partial
from string import Template
from typing import Any, Callable, Dict, List, Union

from drb.drivers.file.file import DrbFileNode
from drb.topics import resolver

from delta.manifest.parser import (
    Dependency,
    Input,
    Manifest,
    Model,
    Output,
    OutputModel,
    Resource,
)
from delta.run.api.model import (
    DataParameterModel,
    ParameterModel,
    PrimitiveParameterModel,
    SecretParameterModel,
    RunContextModel,
    RunContextUpdateModel,
    RunStatus,
    model_param_to_orm_param,
    retrieve_primitive_type,
)
from delta.run.db.orm import ParameterKind, RunContext
from delta.run.job import Image, Job, JobService, PathMapper
from delta.run.job.builder.python_builder import PythonRunner
from delta.run.storage_manager import (
    DeltaLocalStorageManager,
    DeltaStorageManager,
)

logger = logging.getLogger("Orchestrator")
logger.setLevel(logging.DEBUG)


class PlaceholderTemplate(Template):
    idpattern = r"\(([a-z][A-Za-z0-9_-]*)(\.[a-z][A-Za-z0-9_-]*)*\)"


@dataclass
class Edge:
    id: int
    from_id: int
    from_port: str
    to_id: int
    to_port: str
    value: (DataParameterModel | PrimitiveParameterModel |
            SecretParameterModel) = None


@dataclass
class Node:
    id: str
    name: str
    element: Union[Resource, Input, Model, Output, Dependency]
    status: RunStatus = field(init=False, default=RunStatus.CREATED)
    inputs_edges: list[Edge] = field(default_factory=list)
    outputs_edges: list[Edge] = field(default_factory=list)


class DeltaOrchestrator:
    def __init__(
        self,
        manifest: Manifest,
        workflow: dict,
        run_ctx: RunContextModel,
        storage_manager: DeltaStorageManager,
        job_service: JobService,
        on_update_context: Callable[[str, RunContextUpdateModel], None] = None,
        on_create_sub_context: Callable[[str, RunContext], None] = None,
        notify_update_status: Callable[[str, str, RunStatus], None] = None,
    ):
        self._manifest = manifest
        self._workflow = workflow
        self.run_ctx = run_ctx
        self.storage_manager = storage_manager
        self.job_service = job_service
        self._on_update_run_context = on_update_context
        self._on_create_sub_run_context = on_create_sub_context
        self._notify_update_status = notify_update_status
        self.graph = self._init_graph(workflow["edges"])
        self.nodes = self._init_nodes(manifest, workflow)
        self.number_of_nodes = len(self.nodes)
        self._in_degree = self._init_in_degrees(self.graph, self.nodes)

    @staticmethod
    def _init_nodes(manifest: Manifest, workflow: dict):
        # on parcours la liste des edges pour trouver les ports
        nodes = {}
        inputs = collections.defaultdict(list)
        outputs = collections.defaultdict(list)

        for i, edge in enumerate(workflow["edges"]):
            from_id = edge["from"]["id"]
            from_port = edge["from"].get("port", "data")
            to_id = edge["to"]["id"]
            to_port = edge["to"].get("port", "data")
            edge = Edge(i, from_id, from_port, to_id, to_port, None)
            outputs[from_id].append(edge)
            inputs[to_id].append(edge)
        # parcours la liste de noeud
        for node in workflow["nodes"]:
            _id = node["id"]
            (section, name) = node["ref"].split(".", maxsplit=1)
            element = eval(f'manifest.{section}["{name}"]')
            n = Node(_id, name, element, inputs[_id], outputs[_id])
            nodes[_id] = n
        return nodes

    @staticmethod
    def _init_in_degrees(
        graph: Dict[Any, list], nodes: Dict[Any, Node]
    ) -> Dict[Any, int]:
        in_degree = {}
        for node in nodes:
            in_degree[node] = 0
        for i in graph:
            for j in graph[i]:
                in_degree[j] += 1
        return in_degree

    @staticmethod
    def _init_graph(edges: List[dict]) -> Dict[Any, List]:
        graph = collections.defaultdict(list)
        for edge in edges:
            from_id = edge["from"]["id"]
            to_id = edge["to"]["id"]

            graph[from_id].append(to_id)
        return graph

    @staticmethod
    def substitute_output_placeholder(args: dict, output: OutputModel):
        t = PlaceholderTemplate(output.glob)
        sbt_dict = {}
        for id in t.get_identifiers():
            keys = id[1:-1].split(".")

            val = args[keys[0]][keys[1]]
            if isinstance(val, SecretParameterModel):
                val = val.secret_value
            if isinstance(val, PrimitiveParameterModel):
                val = val.value
            if isinstance(val, DataParameterModel):
                val = os.path.basename(val.path)
            sbt_dict[id] = val
        return t.substitute(sbt_dict)

    async def run(self):
        self.storage_manager.add_run_directory(uuid.UUID(self.run_ctx.id))
        self.run_ctx.status = RunStatus.RUNNING
        logger.info(
            "Run(%s) starting (dt: %s, dtv: %s, owner: %s)",
            self.run_ctx.id,
            str(self.run_ctx.deltatwin_id),
            self.run_ctx.deltatwin_version,
            self.run_ctx.owner,
        )

        # Update run context --> set status to running
        rcup = RunContextUpdateModel(status=self.run_ctx.status)
        if self._on_update_run_context:
            self._on_update_run_context(self.run_ctx.id, rcup)
        if self._notify_update_status:
            await self._notify_update_status(
                self.run_ctx.owner,
                self.run_ctx.deltatwin_id,
                self.run_ctx.id,
                self.run_ctx.status,
                self.run_ctx.deltatwin_version,
            )

        try:
            self.check_inputs()
            async with asyncio.TaskGroup() as group:
                for node_id, value in self._in_degree.items():
                    if value == 0:
                        group.create_task(
                            self.exec_task(self.nodes[node_id], group),
                            name=node_id,
                        )
        except TypeError as eg:
            self.run_ctx.status = RunStatus.ERROR
            logger.error(eg)
        except BaseExceptionGroup as eg:
            for ex in eg.exceptions:
                logger.error(ex)
            self.run_ctx.status = RunStatus.ERROR

        if self.run_ctx.status == RunStatus.RUNNING:
            self.run_ctx.status = RunStatus.SUCCESS
            self.run_ctx.return_code = 0

        # update run context --> run completion
        rcup = RunContextUpdateModel(
            status=self.run_ctx.status,
            return_code=self.run_ctx.return_code,
            message=self.run_ctx.message,
            inputs=self.run_ctx.inputs,
            outputs=self.run_ctx.outputs,
        )
        if self._on_update_run_context:
            self._on_update_run_context(self.run_ctx.id, rcup)
        if self._notify_update_status:
            await self._notify_update_status(
                self.run_ctx.owner,
                self.run_ctx.deltatwin_id,
                self.run_ctx.id,
                self.run_ctx.status,
                self.run_ctx.deltatwin_version,
            )

        # logging run completion
        logger.info(
            "Run %s (%s | %s) finished with status : %s (%d: %s)",
            self.run_ctx.id,
            self.run_ctx.deltatwin_id,
            self.run_ctx.deltatwin_version,
            self.run_ctx.status.value,
            self.run_ctx.return_code,
            self.run_ctx.message,
        )
        date_format = "%Y-%m-%d %H:%M:%S"
        started = self.run_ctx.date_created
        computation_time = (
            datetime.datetime.now(datetime.timezone.utc)
            - self.run_ctx.date_created
        )
        ended = started + computation_time
        logger.info(
            "[user-id: %s - process-id: %s - deltatwin-id: %s - "
            "deltatwin-version: %s - deltatwin-status: %s - "
            "starting-date: %s - ending-date: %s]",
            self.run_ctx.owner,
            "run",
            self.run_ctx.deltatwin_id,
            self.run_ctx.deltatwin_version,
            self.run_ctx.status.value,
            started.strftime(date_format),
            ended.strftime(date_format),
        )
        return self.run_ctx

    async def exec_task(self, node: Node, group: asyncio.TaskGroup):
        logger.debug(f"execute run: {self.run_ctx.id}, node: {node.id}")
        node.status = RunStatus.RUNNING

        if isinstance(node.element, Resource):
            await self.execute_resource_node(node, group)
        if isinstance(node.element, Input):
            await self.execute_input_node(node, group)
        if isinstance(node.element, Model):
            await self.execute_model_node(node, group)
        if isinstance(node.element, Output):
            await self.execute_output_node(node, group)
        if isinstance(node.element, Dependency):
            await self.execute_dependency_node(node, group)

    async def fetch_resource(
        self, node: Node, resource: ParameterModel, group: asyncio.TaskGroup
    ):
        logger.debug(f"type: {node.element.type}")
        if node.element.type == "Data" and resource.path is None:
            logger.debug(
                f"node {node.name} is of type Data, fetching value..."
            )
            resource_path = resource.url.strip()
            try:
                start = timeit.default_timer()
                n = resolver.create(resource_path)
                total_time = timeit.default_timer() - start
            except Exception:
                logger.error(f"Error while fetching {node.name}")
                node.status = RunStatus.ERROR
                self.run_ctx.return_code = 210
                self.run_ctx.message = (
                    f"Error while fetching input {node.name} "
                    f"from url {resource_path}. Please verify the url."
                )
                raise RuntimeError(f"Error while fetching {node.name}")
            logger.info(
                "resolved %s, name : %s (in %f s)",
                resource_path,
                n.name,
                total_time,
            )

            start = timeit.default_timer()
            try:
                if isinstance(n, DrbFileNode) and isinstance(
                    self.storage_manager, DeltaLocalStorageManager
                ):
                    logger.debug("Avoid local copy")
                    path = str(n.path)
                else:
                    stream = n.get_impl(io.BufferedIOBase)
                    path = await asyncio.to_thread(
                        self.storage_manager.set_data,
                        self.run_ctx.id,
                        str(node.id),
                        n.name,
                        stream,
                    )
            except Exception as e:
                print(e)
                logger.error(f"Error while fetching {node.name}")
                node.status = RunStatus.ERROR
                self.run_ctx.return_code = 211
                self.run_ctx.message = f"Error while storing input {node.name}"
                raise RuntimeError(f"Error while fetching {node.name}")
            resource.path = path
            infos = await asyncio.to_thread(
                self.storage_manager.get_data_info, path
            )
            resource.size = infos["size"]
            resource.checksum = infos["checksum"]
            total_time = timeit.default_timer() - start
            logger.debug(
                f"file (basename:{os.path.basename(resource.path)},"
                f"size:{resource.size}, "
                f"checksum:{resource.checksum}) "
                f"in {total_time}"
            )

        logger.debug(f"putting value {resource}")
        # then we assign it to the output
        for output_edge in node.outputs_edges:
            output_edge.value = resource
        node.status = RunStatus.SUCCESS

    async def execute_input_node(self, node: Node, group: asyncio.TaskGroup):
        logger.debug(f"executing input node {node.name}")
        # Check if the user specified the input
        _input = None
        for user_input in self.run_ctx.inputs:
            if user_input.name == node.name:
                _input = user_input
        # If not, we get the input from the manifest file
        if _input is None:
            logger.debug("Getting input from manifest")
            input_value = node.element.value
            if node.element.type == "Data":
                _input = DataParameterModel(
                    name=node.name, type="Data", url=input_value["url"]
                )
            elif node.element.type == "secret":
                print(f"input: {input_value}")
                _input = SecretParameterModel(
                    name=node.name,
                    type="secret",
                    value=input_value,
                )
            else:
                _input = PrimitiveParameterModel(
                    name=node.name,
                    type=retrieve_primitive_type(input_value),
                    value=input_value,
                )
            self.run_ctx.inputs.append(_input)

        await self.fetch_resource(node, _input, group)
        rcup = RunContextUpdateModel(inputs=self.run_ctx.inputs)
        if self._on_update_run_context:
            self._on_update_run_context(self.run_ctx.id, rcup)
        self.task_done(node, group)

    async def execute_resource_node(
        self, node: Node, group: asyncio.TaskGroup
    ):
        logger.debug(f"executing resource node {node.name}")
        resource_value = node.element.value
        if node.element.type == "Data":
            resource = DataParameterModel(
                name=node.name, type="Data", url=resource_value
            )
        elif node.element.type == "secret":
            resource = SecretParameterModel(
                name=node.name,
                type="secret",
                value=resource_value,
            )
        else:
            resource = PrimitiveParameterModel(
                name=node.name,
                type=retrieve_primitive_type(resource_value),
                value=resource_value,
            )
        await self.fetch_resource(node, resource, group)
        self.task_done(node, group)

    async def execute_model_node(self, node: Node, group: asyncio.TaskGroup):
        tag = f"{self.run_ctx.deltatwin_id}/{node.name}"
        if self.run_ctx.deltatwin_version:
            tag += f":{self.run_ctx.deltatwin_version}"

        img = Image(tag)
        pr = PythonRunner(node.element)
        args = {}
        inputs = {}
        inputs_mount: List[PathMapper] = []
        outputs_mount: List[PathMapper] = []
        for edge in node.inputs_edges:
            inputs[edge.to_port] = edge.value
            if isinstance(edge.value, DataParameterModel):
                path = self.storage_manager.get_full_path(edge.value.path)
                pm = PathMapper(path, os.path.join("/s3", edge.value.path))
                inputs_mount.append(pm)
        args["inputs"] = inputs
        for output in node.element.outputs.values():
            if output.type == "Data":
                output_path = self.substitute_output_placeholder(args, output)
                path = os.path.join("runs", str(self.run_ctx.id), str(node.id))
                # need "/" at the end for ansible playbook
                # to work with directory
                storage_path = self.storage_manager.get_full_path(path) + "/"
                if not os.path.isabs(output_path):
                    mount_path = os.path.join(".", output_path)
                    pm = PathMapper(mount_path, storage_path)
                else:
                    pm = PathMapper(output_path, storage_path)
                outputs_mount.append(pm)

        cmd = pr.build_command(**args)
        cmd_secure = pr.build_command(secure=True, **args)
        job = Job(
            command=cmd,
            image=img,
            inputs=inputs_mount,
            outputs=outputs_mount,
        )
        logger.debug("Job id : %s", job.id)
        event = asyncio.Event()
        model_done = partial(self.model_done_callback, node, event)
        job.add_callback(model_done)
        await self.job_service.execute_job(job)
        await event.wait()
        if node.status == RunStatus.SUCCESS:
            for output in node.element.outputs.values():
                if output.type == "Data":
                    output_name = self.substitute_output_placeholder(
                        args, output
                    )
                    output_path = self.storage_manager.find_data(
                        self.run_ctx.id, node.id, output_name
                    )
                    if len(output_path) != 1:
                        self.run_ctx.status = RunStatus.ERROR
                        self.run_ctx.return_code = 221
                        self.run_ctx.message = (
                            "Something went wrong during the execution of "
                            f"model {node.name}. Should have only 1 output.\n"
                            f"Node Id: {node.id}\n"
                            f"Image: {tag}\n"
                            f"cmd: {cmd_secure}\n"
                        )
                        return
                    output_path = output_path[0]
                    storage_path = self.storage_manager.get_full_path(
                        output_path
                    )
                    # update output edges with final output name
                    for oe in node.outputs_edges:
                        if oe.from_port == output.name:
                            oe.value = DataParameterModel(
                                name=output.name,
                                type="Data",
                                url=storage_path,
                                path=output_path,
                            )

            self.task_done(node, group)

        elif node.status == RunStatus.ERROR:
            self.run_ctx.status = RunStatus.ERROR
            self.run_ctx.return_code = 221
            self.run_ctx.message = (
                "Something went wrong during the execution"
                f" of model {node.name}.\n"
                f"Node Id: {node.id}\n"
                f"Image: {tag}\n"
                f"cmd: {cmd_secure}\n"
            )

    async def execute_output_node(self, node: Node, group: asyncio.TaskGroup):
        for ie in node.inputs_edges:
            if isinstance(ie.value, DataParameterModel):
                infos = await asyncio.to_thread(
                    self.storage_manager.get_data_info, ie.value.path
                )
                ie.value.size = infos["size"]
                ie.value.checksum = infos["checksum"]
                output = ie.value
                output.name = node.name
            self.run_ctx.outputs.append(output)

        node.status = RunStatus.SUCCESS
        rcup = RunContextUpdateModel(outputs=self.run_ctx.outputs)
        if self._on_update_run_context:
            self._on_update_run_context(self.run_ctx.id, rcup)
        self.task_done(node, group)

    async def execute_dependency_node(
        self, node: Node, group: asyncio.TaskGroup
    ):
        # create a new run context for the dependency
        inputs = []
        # the name of the Data or primitive Parameter of the edges
        # is the name of the output of the previous node
        # so we need to change the name to match
        # the name of the dependency's input
        for ie in node.inputs_edges:
            _input = ie.value
            _input.name = ie.to_port
            inputs.append(_input)
        run_ctx = RunContext(
            id=str(uuid.uuid4()),
            deltatwin_id=node.element.id,
            deltatwin_version=node.element.version,
            owner=self.run_ctx.owner,
            status=RunStatus.CREATED,
            date_created=self.run_ctx.date_created,
            parent_run=self.run_ctx.id,
            inputs=[
                model_param_to_orm_param(p, ParameterKind.INPUT)
                for p in inputs
            ],
            return_code=0,
            message="",
        )

        run = RunContextModel.from_run_context(run_ctx)
        if self._on_create_sub_run_context:
            self._on_create_sub_run_context(self.run_ctx.id, run_ctx)
        logger.debug(
            "Executing dependecy : %s with run id : %s", node.name, run.id
        )

        orch = DeltaOrchestrator(
            self.storage_manager.get_deltatwin_manifest(
                run.deltatwin_id, run.deltatwin_version
            ),
            self.storage_manager.get_deltatwin_workflow(
                run.deltatwin_id, run.deltatwin_version
            ),
            run,
            self.storage_manager,
            self.job_service,
            on_update_context=self._on_update_run_context,
            on_create_sub_context=self._on_create_sub_run_context,
        )
        await orch.run()
        logger.debug(
            "Dependency %s finished with status %s", node.name, run.status
        )
        # run outputs should correspond to the outputs_edges of the node
        if run.status != RunStatus.SUCCESS:
            self.run_ctx.status = RunStatus.ERROR
            self.run_ctx.return_code = 230
            self.run_ctx.message = (
                f"Error while running dependency {node.name}."
                " See dependency error for more info "
                f"(Dependency run id : {run.id})"
            )
            return

        for oe in node.outputs_edges:
            oe.value = next(
                filter(
                    lambda output: output.name == oe.from_port, run.outputs
                ),
                None,
            )
        self.task_done(node, group)

    def model_done_callback(
        self, node: Node, event: asyncio.Event, old_status, new_status
    ):
        node.status = new_status
        event.set()

    def task_done(self, node: Node, group: asyncio.TaskGroup):
        node.status = RunStatus.SUCCESS
        self._in_degree[node.id] -= 1
        for i in self.graph[node.id]:
            self._in_degree[i] -= 1
            if self._in_degree[i] == 0:
                group.create_task(
                    self.exec_task(self.nodes[i], group),
                    name=self.nodes[i].id,
                )

    def check_inputs(self):
        user_inputs = {i.name: i for i in self.run_ctx.inputs}
        missing_inputs = []
        for k, v in self._manifest.inputs.items():
            if v.value is None and k not in user_inputs:
                missing_inputs.append(k)
            if k in user_inputs:
                if v.type == "Data" and not isinstance(
                    user_inputs[k], DataParameterModel
                ):
                    self.run_ctx.return_code = 110
                    if isinstance(
                            user_inputs[k], SecretParameterModel
                    ):
                        self.run_ctx.message = (
                            f"Input {k} must be of type {v.type} "
                            f"but is of type secret."
                        )
                        raise TypeError(
                            (
                                f"Input {k} must be of type {v.type} "
                                "but is of type secret."
                            )
                        )
                    elif isinstance(
                            user_inputs[k], PrimitiveParameterModel
                    ):
                        self.run_ctx.message = (
                            f"Input {k} must be of type {v.type} "
                            f"but is of type "
                            f"{retrieve_primitive_type(user_inputs[k].value)}."
                        )
                        raise TypeError(
                            f"Input {k} must be of type {v.type} "
                            "but is of type "
                            f"{retrieve_primitive_type(user_inputs[k].value)}."
                        )
                elif v.type == "secret" and not isinstance(
                    user_inputs[k], SecretParameterModel
                ):
                    self.run_ctx.return_code = 110
                    if isinstance(
                            user_inputs[k], DataParameterModel
                    ):
                        self.run_ctx.message = (
                            f"Input {k} must be of type {v.type} "
                            f"but is of type Data."
                        )
                        raise TypeError(
                            f"Input {k} must be of type {v.type} "
                            "but is of type Data."
                        )
                    elif isinstance(
                            user_inputs[k], PrimitiveParameterModel
                    ):
                        self.run_ctx.message = (
                            f"Input {k} must be of type {v.type} "
                            f"but is of type "
                            f"{retrieve_primitive_type(user_inputs[k].value)}."
                        )
                        raise TypeError(
                            f"Input {k} must be of type {v.type} "
                            "but is of type "
                            f"{retrieve_primitive_type(user_inputs[k].value)}."
                        )
                elif (v.type != "Data" and v.type != "secret" and
                      v.type != retrieve_primitive_type(user_inputs[k].value)):
                    self.run_ctx.return_code = 110
                    self.run_ctx.message = (
                        f"Input {k} must be of type {v.type} but is of type"
                        f" {retrieve_primitive_type(user_inputs[k].value)}."
                    )
                    raise TypeError(
                        f"Input {k} must be of type {v.type} "
                        "but is of type "
                        f"{retrieve_primitive_type(user_inputs[k].value)}."
                    )
        if missing_inputs:
            self.run_ctx.return_code = 120
            self.run_ctx.message = f"Inputs {missing_inputs} required."
            raise TypeError(f"Inputs {missing_inputs} required.")

    def are_user_inputs_valid(self):
        is_valid = True
        manifest = self.storage_manager.get_deltatwin_manifest(
            self.run_ctx.deltatwin_id, self.run_ctx.deltatwin_version
        )
        for i in self.run_ctx.inputs:
            input = manifest.inputs.get(i.name, None)
            if input is None:
                logger.error(f"Input {i} is not a valid input")
                is_valid = False
            type = input.get("type", None)
            if type is None:
                logger.error(f"Input {i} has no type")
                is_valid = False
        return is_valid

    def is_valid_dag(self):
        top_sorted = []
        queue = []
        in_degree = self._init_in_degrees(self.graph, self.nodes)
        for i in self.graph:
            if in_degree[i] == 0:
                queue.append(i)

        cnt = 0
        while queue:
            u = queue.pop(0)
            top_sorted.append(u)
            for i in self.graph[u]:
                in_degree[i] -= 1
                if in_degree[i] == 0:
                    queue.append(i)

            cnt += 1

        # Check if there was a cycle
        if cnt != self.number_of_nodes:
            return False
        logger.debug("Possible node execution order: %s", top_sorted)
        return True
